# Copyright (c) 2023 BeamNG GmbH, Angelo Matteo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import bpy
import blf
import bmesh
import base64
import pickle
import sys # <<< Added import
import traceback # <<< Added import

from blf import position as blfpos
from blf import size as blfsize
from blf import draw as blfdraw
from blf import color as blfcolor
from blf import dimensions as blfdims

from bpy_extras.view3d_utils import location_3d_to_region_2d
from mathutils import Vector, Matrix

# Import from local modules
from . import constants
from . import globals as jb_globals # Import globals
from . import text_editor
from . import sjsonast
from . import bng_sjson
from .text_editor import SCENE_SHORT_TO_FULL_FILENAME
from .sjsonast import ASTNode # <<< Ensure ASTNode is imported
from .jbeam import io as jbeam_io # <<< ADDED: Import jbeam_io

if not constants.UNIT_TESTING:
    import gpu
    from gpu_extras.batch import batch_for_shader

# Drawing related globals
veh_render_dirty = False # Flag to trigger batch rebuild
part_name_to_obj: dict[str, bpy.types.Object] = {} # Cache for vehicle part objects
warned_missing_nodes_this_rebuild = set() # <<< MOVED: Make this a module-level variable

# --- Visualization Batches & Coords ---
beam_render_shader = None
beam_render_batch = None # NORMAL beams (Green)
beam_coords = []

anisotropic_beam_render_batch = None # Magenta Lines (Default) <<< CORRECTED COMMENT
anisotropic_beam_coords = []

support_beam_render_batch = None
support_beam_coords = []

hydro_beam_render_batch = None
hydro_beam_coords = []

bounded_beam_render_batch = None
bounded_beam_coords = []

lbeam_render_batch = None
lbeam_coords = []

pressured_beam_render_batch = None
pressured_beam_coords = []

torsionbar_render_batch = None
torsionbar_coords = []
torsionbar_red_render_batch = None
torsionbar_red_coords = []

rail_render_batch = None
rail_coords = []

# --- Cross-Part Beam Visualization ---
cross_part_beam_render_batch = None
cross_part_beam_coords = []
all_nodes_cache: dict[str, tuple[Vector, str, str]] = {} # {node_id: (world_pos, source_filepath, part_origin)}
all_nodes_cache_dirty = True # Flag to rebuild cache

# --- Highlight on Click ---
highlight_render_batch = None # For non-torsionbar elements
highlight_coords = [] # For non-torsionbar elements
highlight_torsionbar_outer_batch = None
highlight_torsionbar_outer_coords = []
highlight_torsionbar_mid_batch = None
highlight_torsionbar_mid_coords = []
# highlighted_node_ids (set) and highlighted_element_ordered_node_ids (list) are managed in globals.py

# Update function for the cross-part beam visibility toggle.
def _update_toggle_cross_part_beams_vis(self, context):
    scene = context.scene
    # Always trigger a redraw/rebuild when the toggle changes
    scene.jbeam_editor_veh_render_dirty = True # Use scene property

# Helper function to find the line number of a beam in the AST
def find_beam_line_number(jbeam_filepath: str, target_part_origin: str, target_id1: str, target_id2: str):
    """
    Finds the 1-based line number of a specific beam definition in a JBeam file
    by matching node IDs.
    """
    # <<< ADDED CHECK >>>
    # Ignore beams involving transient TEMP_ IDs during search
    if target_id1.startswith('TEMP_') or target_id2.startswith('TEMP_'):
        return None
    # <<< END ADDED CHECK >>>

    file_content = text_editor.read_int_file(jbeam_filepath)
    if not file_content:
        print(f"Error: Could not read internal file: {jbeam_filepath}", file=sys.stderr)
        return None

    try:
        ast_data = sjsonast.parse(file_content)
        if not ast_data:
            print(f"Error: Could not parse AST for: {jbeam_filepath}", file=sys.stderr)
            return None

        ast_nodes = ast_data['ast']['nodes']
        sjsonast.calculate_char_positions(ast_nodes)

        stack = []
        in_dict = True
        pos_in_arr = 0
        temp_dict_key = None
        dict_key = None

        i = 0
        while i < len(ast_nodes):
            node: sjsonast.ASTNode = ast_nodes[i]
            node_type = node.data_type

            if node_type == 'wsc':
                i += 1
                continue

            # --- Determine context based on stack ---
            in_target_part = len(stack) > 0 and stack[0][0] == target_part_origin
            in_beams_section = (
                in_target_part and
                len(stack) == 2 and
                stack[1][0] == 'beams' and
                not in_dict # Ensure we are inside the beams *array*
            )
            # --- End context determination ---

            if in_dict:
                if node_type == '{':
                    if dict_key is not None:
                        stack.append((dict_key, True)) # Parent was dict
                    pos_in_arr = 0; temp_dict_key = None; dict_key = None; in_dict = True
                elif node_type == '[':
                    if dict_key is not None:
                        stack.append((dict_key, True)) # Parent was dict
                    pos_in_arr = 0; temp_dict_key = None; dict_key = None; in_dict = False
                elif node_type == '}':
                    if stack:
                        prev_key_or_idx, prev_in_dict = stack.pop()
                        in_dict = prev_in_dict
                        pos_in_arr = prev_key_or_idx + 1 if not prev_in_dict else 0
                    else: in_dict = None
                elif node_type == ']':
                     print(f"Error: Unexpected ']' while expecting dict elements near pos {node.start_pos}", file=sys.stderr); return None
                else:
                    if temp_dict_key is None and node_type == '"': temp_dict_key = node.value
                    elif node_type == ':': dict_key = temp_dict_key
                    elif dict_key is not None: temp_dict_key = None; dict_key = None
            else: # In array
                if node_type == '[': # Start of a beam definition array
                    beam_entry_start_node = node
                    stack.append((pos_in_arr, False)) # Parent was array
                    pos_in_arr = 0; in_dict = False

                    if in_beams_section: # Use derived flag
                        found_id1 = None; found_id2 = None
                        k = i + 1; ids_found = 0
                        while k < len(ast_nodes):
                            inner_node = ast_nodes[k]
                            if inner_node.data_type == ']': break
                            if inner_node.data_type == '"':
                                ids_found += 1
                                if ids_found == 1: found_id1 = inner_node.value
                                elif ids_found == 2: found_id2 = inner_node.value; break
                            k += 1

                        if found_id1 is not None and found_id2 is not None:
                            if (found_id1 == target_id1 and found_id2 == target_id2) or \
                               (found_id1 == target_id2 and found_id2 == target_id1):
                                start_char_pos = beam_entry_start_node.start_pos
                                line_number = file_content[:start_char_pos].count('\n') + 1
                                return line_number
                elif node_type == '{':
                    stack.append((pos_in_arr, False)) # Parent was array
                    pos_in_arr = 0; in_dict = True
                elif node_type == ']':
                    if stack:
                        prev_key_or_idx, prev_in_dict = stack.pop()
                        in_dict = prev_in_dict
                        pos_in_arr = prev_key_or_idx + 1 if not prev_in_dict else 0
                    else: in_dict = None
                elif node_type == '}':
                    print(f"Error: Unexpected '}}' while expecting array elements near pos {node.start_pos}", file=sys.stderr); return None
                else: # Value node within the current array
                    pos_in_arr += 1 # Increment position *after* processing the current element
            i += 1

        # Don't print a warning here, as TEMP_ nodes will naturally not be found
        # print(f"Warning: Beam {target_id1}-{target_id2} not found in part '{target_part_origin}' in file {jbeam_filepath}", file=sys.stderr)
        return None

    except Exception as e:
        print(f"Error finding beam line number: {e}", file=sys.stderr)
        traceback.print_exc()
        return None

# Helper function to find the line number of a node in the AST (REVISED APPROACH)
def find_node_line_number(jbeam_filepath: str, target_part_origin: str, target_node_id: str):
    """
    Finds the 1-based line number of a specific node definition in a JBeam file.
    (Revised approach focusing directly on the target part's nodes section)
    """
    # <<< ADDED CHECK >>>
    # Ignore transient TEMP_ IDs during search
    if target_node_id.startswith('TEMP_'):
        return None
    # <<< END ADDED CHECK >>>

    file_content = text_editor.read_int_file(jbeam_filepath)
    if not file_content:
        print(f"Error: Could not read internal file: {jbeam_filepath}", file=sys.stderr)
        return None

    try:
        # Use the robust bng_sjson parser first to get the structure
        # Add padding for safety
        padded_content = file_content + chr(127) * 2
        c, i = bng_sjson._skip_white_space(padded_content, 0, jbeam_filepath)
        parsed_data = None
        if c == 123:
            parsed_data, _ = bng_sjson._read_object(padded_content, i, jbeam_filepath)
        else:
            print(f"Error: JBeam file does not start with '{{': {jbeam_filepath}", file=sys.stderr)
            return None

        if not parsed_data or target_part_origin not in parsed_data:
            # Don't warn if the part itself isn't found (might happen during load/revert)
            # print(f"Warning: Part '{target_part_origin}' not found in parsed data for {jbeam_filepath}", file=sys.stderr)
            return None

        part_data = parsed_data[target_part_origin]
        if not isinstance(part_data, dict) or 'nodes' not in part_data:
            # print(f"Warning: 'nodes' section not found in part '{target_part_origin}' in {jbeam_filepath}", file=sys.stderr)
            return None

        nodes_section = part_data['nodes']
        if not isinstance(nodes_section, list) or len(nodes_section) <= 1:
            # print(f"Warning: 'nodes' section in part '{target_part_origin}' is not a valid list or is empty/header-only in {jbeam_filepath}", file=sys.stderr)
            return None

        # Now parse with AST to get line numbers
        ast_data = sjsonast.parse(file_content)
        if not ast_data:
            print(f"Error: Could not parse AST for: {jbeam_filepath}", file=sys.stderr)
            return None

        ast_nodes = ast_data['ast']['nodes']
        sjsonast.calculate_char_positions(ast_nodes)

        # --- Find the AST nodes corresponding to the target part's 'nodes' array ---
        nodes_array_start_node_idx = -1
        nodes_array_end_node_idx = -1
        nodes_array_stack_depth = -1 # <<< Store stack depth when '[' is found

        stack = []
        in_dict = True
        pos_in_arr = 0
        temp_dict_key = None
        dict_key = None
        in_target_part = False # <<< Track if we are inside the target part

        i = 0
        while i < len(ast_nodes):
            node: sjsonast.ASTNode = ast_nodes[i]
            node_type = node.data_type

            if node_type == 'wsc':
                i += 1
                continue

            # --- Track if inside target part ---
            current_stack_depth = len(stack)
            if current_stack_depth == 1 and stack[0][0] == target_part_origin:
                in_target_part = True
            elif current_stack_depth < 1:
                in_target_part = False
                # If we exited the target part after finding the start, stop searching
                if nodes_array_start_node_idx != -1:
                    # print(f"Warning: Exited target part '{target_part_origin}' before finding end of 'nodes' array.", file=sys.stderr)
                    break # Optimization: stop if we leave the part
            # --- End tracking ---

            if in_dict:
                if node_type == '{':
                    if dict_key is not None:
                        stack.append((dict_key, True))
                    pos_in_arr = 0; temp_dict_key = None; dict_key = None; in_dict = True
                elif node_type == '[':
                    if dict_key is not None:
                        stack.append((dict_key, True))
                        # Check if we are entering the 'nodes' array within the target part
                        if in_target_part and dict_key == 'nodes':
                            nodes_array_start_node_idx = i
                            nodes_array_stack_depth = len(stack) # <<< Record stack depth
                    pos_in_arr = 0; temp_dict_key = None; dict_key = None; in_dict = False
                elif node_type == '}':
                    if stack:
                        prev_key, prev_in_dict = stack.pop()
                        in_dict = prev_in_dict
                        pos_in_arr = prev_key + 1 if not prev_in_dict else 0
                    else: in_dict = None
                elif node_type == ']':
                     print(f"Error: Unexpected ']' while expecting dict elements near pos {node.start_pos}", file=sys.stderr); return None
                else:
                    if temp_dict_key is None and node_type == '"': temp_dict_key = node.value
                    elif node_type == ':': dict_key = temp_dict_key
                    elif dict_key is not None: temp_dict_key = None; dict_key = None
            else: # In array
                if node_type == '[':
                    stack.append((pos_in_arr, False))
                    pos_in_arr = 0; in_dict = False
                elif node_type == '{':
                    stack.append((pos_in_arr, False))
                    pos_in_arr = 0; in_dict = True
                elif node_type == ']':
                    # <<< Revised End Bracket Logic >>>
                    # Check if we found the start and if the current stack depth matches the recorded depth
                    # The stack length *before* popping should match the recorded depth.
                    if nodes_array_start_node_idx != -1 and len(stack) == nodes_array_stack_depth:
                        nodes_array_end_node_idx = i
                        break # Found the end, exit the main loop
                    # <<< End Revised Logic >>>

                    # Normal stack pop logic if it wasn't the target array's end
                    if stack:
                        prev_key_or_idx, prev_in_dict = stack.pop()
                        in_dict = prev_in_dict
                        pos_in_arr = prev_key_or_idx + 1 if not prev_in_dict else 0
                    else: in_dict = None
                elif node_type == '}':
                    print(f"Error: Unexpected '}}' while expecting array elements near pos {node.start_pos}", file=sys.stderr); return None
                else:
                    pos_in_arr += 1
            i += 1

        # --- Process the identified 'nodes' array in the AST ---
        if nodes_array_start_node_idx == -1 or nodes_array_end_node_idx == -1:
            # print(f"Warning: Could not locate AST boundaries for 'nodes' section in part '{target_part_origin}' in {jbeam_filepath}", file=sys.stderr)
            return None

        node_header = []
        node_id_column_index = -1
        row_start_node_idx = -1

        # Iterate specifically within the nodes array bounds
        k = nodes_array_start_node_idx + 1 # Start after the opening '['
        while k < nodes_array_end_node_idx:
            node = ast_nodes[k]
            node_type = node.data_type

            if node_type == 'wsc':
                k += 1
                continue

            if node_type == '[': # Start of a row
                row_start_node_idx = k
                current_col_index = 0
                is_header_row = (len(node_header) == 0) # Assume first row is header

                # Iterate within the row
                j = k + 1
                while j < nodes_array_end_node_idx:
                    inner_node = ast_nodes[j]
                    inner_node_type = inner_node.data_type

                    if inner_node_type == 'wsc':
                        j += 1
                        continue
                    if inner_node_type == ']': # End of row
                        k = j # Move outer loop index past this row
                        break

                    # Process value node within the row
                    if is_header_row:
                        if inner_node_type == '"':
                            node_header.append(inner_node.value)
                            if inner_node.value == 'id':
                                node_id_column_index = current_col_index
                    else: # Data row
                        if node_id_column_index != -1 and current_col_index == node_id_column_index:
                            if inner_node_type == '"' and inner_node.value == target_node_id:
                                # Found the target node ID in the correct column!
                                start_char_pos = ast_nodes[row_start_node_idx].start_pos
                                line_number = file_content[:start_char_pos].count('\n') + 1
                                return line_number

                    current_col_index += 1
                    j += 1
            k += 1

        # If loop finishes without finding the node
        # Don't print a warning here, as TEMP_ nodes will naturally not be found
        # print(f"Warning: Node ID '{target_node_id}' not found within 'nodes' section of part '{target_part_origin}' in file {jbeam_filepath}", file=sys.stderr)
        return None

    except Exception as e:
        print(f"Error finding node line number: {e}", file=sys.stderr)
        traceback.print_exc()
        return None

# Helper function to scroll Text Editor
def _scroll_editor_to_line(context: bpy.types.Context, filepath: str, line: int):
    """Scrolls the Text Editor to the specified file and line."""
    short_filename = text_editor._to_short_filename(filepath)
    text_obj = bpy.data.texts.get(short_filename)

    if not text_obj:
        print(f"Text object not found: {short_filename}", file=sys.stderr)
        return False

    scrolled = False
    for window in context.window_manager.windows:
        screen = window.screen
        for area in screen.areas:
            if area.type == 'TEXT_EDITOR':
                space = area.spaces[0]
                if space.text == text_obj:
                    line_index = max(0, line - 1)
                    text_obj.cursor_set(line_index)
                    with context.temp_override(window=window, area=area):
                        bpy.ops.text.jump(line=line)
                    area.tag_redraw()
                    scrolled = True
                    break
        if scrolled:
            break

    return scrolled

# Update the cache of all node positions from all loaded JBeam files
def update_all_nodes_cache(context: bpy.types.Context):
    """Scans ALL loaded JBeam text files and caches node positions and part origins."""
    global all_nodes_cache, all_nodes_cache_dirty
    print("Updating all nodes cache...")
    all_nodes_cache.clear()
    scene = context.scene
    ui_props = scene.ui_properties

    short_to_full_map = scene.get(SCENE_SHORT_TO_FULL_FILENAME, {})
    if not short_to_full_map:
        print("Scene mapping not found or empty, cannot update nodes cache yet.")
        all_nodes_cache_dirty = False
        return

    for short_name, text_obj in bpy.data.texts.items():
        full_filepath = short_to_full_map.get(short_name)
        if not full_filepath or not full_filepath.lower().endswith('.jbeam'):
            continue

        try:
            file_content = text_obj.as_string()
            if not file_content or '"nodes"' not in file_content:
                continue

            # Wrap parsing in try-except SyntaxError
            try:
                padded_content = file_content + chr(127) * 2
                c, i = bng_sjson._skip_white_space(padded_content, 0, full_filepath)
                parsed_data = None
                if c == 123:
                    parsed_data, _ = bng_sjson._read_object(padded_content, i, full_filepath)
                else:
                    continue # Skip if file doesn't start with '{' after whitespace/comments

                if not parsed_data: continue

                for part_name, part_data in parsed_data.items():
                    if isinstance(part_data, dict) and 'nodes' in part_data:
                        nodes_section = part_data['nodes']
                        if isinstance(nodes_section, list) and len(nodes_section) > 1:
                            header = nodes_section[0]
                            try:
                                id_idx = header.index("id")
                                x_idx = header.index("posX")
                                y_idx = header.index("posY")
                                z_idx = header.index("posZ")
                            except (ValueError, IndexError):
                                continue # Skip if header is malformed

                            for node_row in nodes_section[1:]:
                                if isinstance(node_row, list) and len(node_row) > max(id_idx, x_idx, y_idx, z_idx):
                                    node_id = node_row[id_idx]
                                    try:
                                        pos_x_val = node_row[x_idx]
                                        pos_y_val = node_row[y_idx]
                                        pos_z_val = node_row[z_idx]
                                        # Check if values are numbers before converting
                                        if isinstance(pos_x_val, (int, float)) and \
                                           isinstance(pos_y_val, (int, float)) and \
                                           isinstance(pos_z_val, (int, float)):
                                            pos = Vector((float(pos_x_val), float(pos_y_val), float(pos_z_val)))
                                            all_nodes_cache[node_id] = (pos, full_filepath, part_name)
                                        # else: # Skip nodes with expression-based positions silently
                                        #    pass
                                    except ValueError:
                                         pass # Skip nodes with invalid number formats silently
                                    except TypeError as e:
                                         print(f"Warning: Could not parse node position for '{node_id}' in {full_filepath} (TypeError): {e}", file=sys.stderr)
            except SyntaxError as se:
                # Print a warning instead of a full traceback for syntax errors during cache update
                print(f"Warning: Skipping node cache update for '{full_filepath}' due to syntax error: {se}", file=sys.stderr)
                continue # Continue to the next file

        except Exception as e:
            # Catch other potential errors during file processing
            print(f"Error processing file {full_filepath} for node cache: {e}", file=sys.stderr)
            traceback.print_exc() # Print traceback for unexpected errors

    print(f"All nodes cache updated with {len(all_nodes_cache)} nodes.")
    all_nodes_cache_dirty = False

# Refresh the current JBeam data based on the active object
def refresh_curr_vdata(force_refresh=False):
    global veh_render_dirty, all_nodes_cache_dirty, warned_missing_nodes_this_rebuild # <<< ADD warned_missing_nodes_this_rebuild
    context = bpy.context
    scene = context.scene
    ui_props = scene.ui_properties

    selected_obj_name = None
    jbeam_part = None
    is_new_jbeam_object = False

    obj = context.active_object
    if obj is not None:
        obj_data = obj.data
        if obj_data and obj_data.get(constants.MESH_JBEAM_PART) is not None:
            jbeam_part = obj_data.get(constants.MESH_JBEAM_PART)
            selected_obj_name = obj.name
            if jb_globals.prev_obj_selected != selected_obj_name:
                 is_new_jbeam_object = True
        else:
            selected_obj_name = None; jbeam_part = None
    else:
        selected_obj_name = None

    object_changed = jb_globals.prev_obj_selected != selected_obj_name

    if force_refresh or object_changed:
        if jbeam_part is not None and obj is not None:
            collection = obj.users_collection[0] if obj.users_collection else None
            veh_model = collection.get(constants.COLLECTION_VEHICLE_MODEL) if collection else None
            try:
                if veh_model is not None and collection.get(constants.COLLECTION_VEHICLE_BUNDLE):
                    jb_globals.curr_vdata = pickle.loads(base64.b64decode(collection[constants.COLLECTION_VEHICLE_BUNDLE]))['vdata']
                elif obj_data.get(constants.MESH_SINGLE_JBEAM_PART_DATA):
                    jb_globals.curr_vdata = pickle.loads(base64.b64decode(obj_data[constants.MESH_SINGLE_JBEAM_PART_DATA]))
                else:
                    jb_globals.curr_vdata = None
            except (TypeError, KeyError, EOFError, pickle.UnpicklingError, base64.binascii.Error) as e:
                 print(f"Error loading JBeam data for {selected_obj_name}: {e}", file=sys.stderr)
                 jb_globals.curr_vdata = None
        else:
            jb_globals.curr_vdata = None

        # Always mark cache dirty if object changed or forced refresh
        if is_new_jbeam_object or force_refresh:
             all_nodes_cache_dirty = True
             warned_missing_nodes_this_rebuild.clear() # <<< ADDED: Clear warnings on significant refresh

        veh_render_dirty = True

        # --- ADDED: Trigger highlight update on object change ---
        # If the object changed, it's a JBeam part, and highlighting is enabled,
        # re-run the highlight logic based on the current text editor state.
        if object_changed and jbeam_part is not None and ui_props.highlight_element_on_click:
            try:
                text_area = None
                # Find the first visible text editor
                if context.window_manager:
                    for window in context.window_manager.windows:
                        screen = window.screen
                        for area in screen.areas:
                            if area.type == "TEXT_EDITOR":
                                text_area = area
                                break
                        if text_area: break

                if text_area:
                    space = text_area.spaces.active
                    if space and space.text:
                        text_obj = space.text
                        current_line_index = text_obj.current_line_index
                        # Call the highlight function directly using the current text editor state
                        # This ensures the highlight reflects the newly selected object's context
                        find_and_highlight_element_for_line(context, text_obj, current_line_index)
                        # find_and_highlight_element_for_line already handles setting veh_render_dirty
                        # and tagging redraw, so no need to do it again here.
            except Exception as e:
                print(f"Error triggering highlight update on object change: {e}", file=sys.stderr)
                traceback.print_exc()
        # --- END ADDED ---

        jb_globals.prev_obj_selected = selected_obj_name


# --- START MOVED HIGHLIGHT LOGIC ---

def _tag_redraw_3d_views(context: bpy.types.Context):
    """Helper function to tag all 3D View areas for redraw."""
    if not context.window_manager: return
    for window in context.window_manager.windows:
        for area in window.screen.areas:
            if area.type == 'VIEW_3D':
                area.tag_redraw()

# <<< START MODIFIED FUNCTION find_and_highlight_element_for_line >>>
def find_and_highlight_element_for_line(context: bpy.types.Context, text_obj: bpy.types.Text, line_index: int):
    """
    Parses the JBeam file content around the given line index,
    identifies the *first* JBeam element (node, beam, rail, torsionbar, slidenode) definition overlapping that line,
    verifies its structural context, finds its coordinates (if applicable), and updates global highlight state.
    Returns True if an element was found and highlighted, False otherwise.
    """
    global veh_render_dirty, part_name_to_obj # Ensure part_name_to_obj is global if modified

    scene = context.scene
    ui_props = scene.ui_properties
    short_to_full_map = scene.get(SCENE_SHORT_TO_FULL_FILENAME, {})
    full_filepath = short_to_full_map.get(text_obj.name)

    # Clear previous highlight coordinates and node IDs first
    highlight_coords.clear()
    highlight_torsionbar_outer_coords.clear()
    highlight_torsionbar_mid_coords.clear()
    jb_globals.highlighted_node_ids.clear() # Clear the set
    jb_globals.highlighted_element_ordered_node_ids.clear() # Clear the ordered list
    # Store previous type before clearing, to check if redraw is needed on failure
    prev_highlight_type = jb_globals.highlighted_element_type
    jb_globals.highlighted_element_type = None # Assume failure until success

    # Update last info immediately, regardless of success/failure below,
    # to prevent re-processing the same line if nothing is found.
    jb_globals.last_text_area_info['name'] = text_obj.name
    jb_globals.last_text_area_info['line_index'] = line_index

    # Default redraw trigger check
    needs_redraw = prev_highlight_type is not None

    if not full_filepath:
        if needs_redraw: veh_render_dirty = True; _tag_redraw_3d_views(context)
        return False

    file_content = text_obj.as_string()
    if not file_content:
        if needs_redraw: veh_render_dirty = True; _tag_redraw_3d_views(context)
        return False

    lines = file_content.splitlines(True) # Keep ends for accurate length
    if line_index >= len(lines):
        if needs_redraw: veh_render_dirty = True; _tag_redraw_3d_views(context)
        return False # Cursor out of bounds

    try:
        # --- Initialize variables ---
        node_ids = [] # This will store the ordered list from parsing (used for most elements)
        slidenode_node_id = None # <<< Specific storage for slidenode
        slidenode_rail_name = None # <<< Specific storage for slidenode
        element_type = None # Determined by AST context
        original_color = (1,1,1,1)
        original_mid_color = (1,0,0,1) # Default mid color
        original_width = 1.0
        beam_type_from_data = '|NORMAL' # Default beam type

        # --- AST Parsing and Context Check ---
        ast_data = sjsonast.parse(file_content)
        if not ast_data:
            if needs_redraw: veh_render_dirty = True; _tag_redraw_3d_views(context)
            return False # Cannot parse AST

        ast_nodes = ast_data['ast']['nodes']
        sjsonast.calculate_char_positions(ast_nodes)

        # Calculate character position range for the target line
        line_start_char_pos = sum(len(l) for l in lines[:line_index])
        line_end_char_pos = line_start_char_pos + len(lines[line_index]) # Exclusive end

        # Traverse AST to find the first element definition on the target line and check context
        stack = []
        in_dict = True
        pos_in_arr = 0
        temp_dict_key = None
        dict_key = None
        current_part_name = None
        current_section_name = None
        found_element_on_line = False
        # <<< NEW: Track rail context >>>
        in_rails_section_dict = False
        current_rail_name = None
        in_rail_links_array = False

        i = 0
        while i < len(ast_nodes):
            node: sjsonast.ASTNode = ast_nodes[i]
            node_type = node.data_type

            # Optimization: If node start is beyond the line end, stop searching
            if node.start_pos >= line_end_char_pos and not found_element_on_line:
                # Only break if we haven't already found the element start
                break

            if node_type == 'wsc':
                i += 1
                continue

            # --- Stack and Context Management ---
            # <<< MODIFIED: More detailed context tracking >>>
            if in_dict:
                if node_type == '{':
                    if dict_key is not None:
                        stack.append((dict_key, True)) # Parent was dict
                        if len(stack) == 1: current_part_name = dict_key
                        if len(stack) == 2 and dict_key == 'rails': in_rails_section_dict = True
                        if in_rails_section_dict and len(stack) == 3: current_rail_name = dict_key # Entering a specific rail's dict
                    dict_key = None; temp_dict_key = None; in_dict = True
                elif node_type == '[':
                    if dict_key is not None:
                        stack.append((dict_key, True)) # Parent was dict
                        if len(stack) == 1: current_part_name = dict_key
                        # <<< MODIFIED: Include 'slidenodes' >>>
                        if len(stack) == 2: current_section_name = dict_key # Entering nodes/beams/torsionbars/slidenodes array
                        # <<< NEW: Check if entering "links:" array >>>
                        if in_rails_section_dict and len(stack) == 4 and dict_key == 'links:':
                            in_rail_links_array = True
                    dict_key = None; temp_dict_key = None; in_dict = False
                elif node_type == '}':
                    if stack:
                        prev_key, prev_in_dict = stack.pop()
                        if len(stack) == 2 and prev_key == current_rail_name: current_rail_name = None # Exiting specific rail dict
                        if len(stack) == 1 and prev_key == 'rails': in_rails_section_dict = False
                        if len(stack) == 0: current_part_name = None
                        in_dict = prev_in_dict
                    else: in_dict = None
                elif node_type == ']':
                     pass # Should not be reached if structure is valid dict
                else:
                    if temp_dict_key is None and node_type == '"': temp_dict_key = node.value
                    elif node_type == ':': dict_key = temp_dict_key
                    elif dict_key is not None: # Value node after key:
                        # --- Check for overlap on value node ---
                        node_overlaps_line = (node.start_pos < line_end_char_pos and node.end_pos >= line_start_char_pos)
                        if node_overlaps_line and in_rails_section_dict and current_rail_name is not None and not found_element_on_line:

                             rail_dict_start_idx = -1
                             temp_k = i - 1 # Start searching backwards from the value node
                             open_brackets = 0
                             while temp_k >= 0:
                                 if ast_nodes[temp_k].data_type == '}': open_brackets += 1
                                 elif ast_nodes[temp_k].data_type == '{':
                                     if open_brackets == 0:
                                         rail_dict_start_idx = temp_k
                                         break
                                     open_brackets -= 1
                                 temp_k -= 1

                             if rail_dict_start_idx != -1:
                                 # Search forward from the rail's '{' for "links:"
                                 links_key_found = False
                                 temp_k = rail_dict_start_idx + 1
                                 while temp_k < len(ast_nodes):
                                     inner_node = ast_nodes[temp_k]
                                     if inner_node.data_type == '"' and inner_node.value == 'links:':
                                         links_key_found = True
                                     elif links_key_found and inner_node.data_type == '[':
                                         # Found the links array, parse it
                                         temp_node_ids = []
                                         l = temp_k + 1
                                         while l < len(ast_nodes):
                                             link_node = ast_nodes[l]
                                             if link_node.data_type == ']': break
                                             if link_node.data_type == '"': temp_node_ids.append(link_node.value)
                                             l += 1
                                         if len(temp_node_ids) == 2:
                                             element_type = 'rail'
                                             node_ids = temp_node_ids
                                             found_element_on_line = True
                                             break # Exit inner search loop
                                         else:
                                             break # Exit inner search loop
                                     elif inner_node.data_type == '}': # Reached end of rail dict
                                         break # Exit inner search loop
                                     temp_k += 1
                             if found_element_on_line: break # Exit outer loop if found

                        # Reset key tracking after processing value
                        dict_key = None; temp_dict_key = None
            else: # In array
                if node_type == '[': # Start of an array element (potential JBeam definition)
                    node_overlaps_line = (node.start_pos < line_end_char_pos and node.end_pos >= line_start_char_pos)

                    # <<< MODIFIED: Check array context, include 'slidenodes' >>>
                    is_element_array = (
                        (current_section_name == 'nodes' and len(stack) == 2) or
                        (current_section_name == 'beams' and len(stack) == 2) or
                        (current_section_name == 'torsionbars' and len(stack) == 2) or
                        (current_section_name == 'slidenodes' and len(stack) == 2) or # <<< ADDED: Check for slidenodes section
                        (in_rail_links_array and len(stack) == 4) # Check if it's the links array itself
                    )

                    if node_overlaps_line and is_element_array and not found_element_on_line:
                        # print(f"[Highlight Debug] '[' at {node.start_pos}-{node.end_pos} overlaps target line. Section: {current_section_name}, Rail Links: {in_rail_links_array}")
                        # Parse content within this bracket pair from AST
                        temp_node_ids = []
                        temp_values_count = 0
                        k = i + 1
                        while k < len(ast_nodes):
                            inner_node = ast_nodes[k]
                            if inner_node.data_type == ']': break
                            if inner_node.data_type == '"': temp_node_ids.append(inner_node.value)
                            elif inner_node.data_type == 'number' or inner_node.data_type == 'bool': temp_values_count += 1
                            elif inner_node.data_type == '{': break # Options dict
                            k += 1

                        num_parsed_ids = len(temp_node_ids)
                        # print(f"[Highlight Debug] Parsed content: ids={temp_node_ids}, values={temp_values_count}")

                        # Determine element type based on context and parsed content
                        if current_section_name == 'nodes' and num_parsed_ids == 1 and temp_values_count >= 3:
                            element_type = 'node'
                            node_ids = temp_node_ids
                            found_element_on_line = True
                        elif current_section_name == 'beams' and num_parsed_ids == 2:
                            element_type = 'beam'
                            node_ids = temp_node_ids
                            found_element_on_line = True
                        # <<< MODIFIED: Check rail context here >>>
                        elif in_rail_links_array and num_parsed_ids == 2:
                            element_type = 'rail'
                            node_ids = temp_node_ids
                            found_element_on_line = True
                        elif current_section_name == 'torsionbars' and num_parsed_ids == 4:
                            element_type = 'torsionbar'
                            node_ids = temp_node_ids
                            found_element_on_line = True
                        # <<< ADDED: Check for slidenodes >>>
                        elif current_section_name == 'slidenodes' and num_parsed_ids >= 2:
                            element_type = 'slidenode'
                            slidenode_node_id = temp_node_ids[0] # Store the node ID
                            slidenode_rail_name = temp_node_ids[1] # Store the rail name
                            found_element_on_line = True
                        # <<< END ADDED >>>

                        if found_element_on_line:
                            # print(f"[Highlight Debug] Element FOUND on line: type={element_type}, ids={node_ids}")
                            break # Found the first relevant element on the line
                        else:
                            # print(f"[Highlight Debug] Parsed array content on line did not match criteria. Section: {current_section_name}, Rail Links: {in_rail_links_array}")
                            pass # Keep searching

                    # Normal stack push if not the target element or context wrong
                    stack.append((pos_in_arr, False)) # Parent was array
                    pos_in_arr = 0; in_dict = False
                elif node_type == '{':
                    stack.append((pos_in_arr, False)) # Parent was array
                    pos_in_arr = 0; in_dict = True
                elif node_type == ']':
                    if stack:
                        prev_key_or_idx, prev_in_dict = stack.pop()
                        # <<< NEW: Check if exiting "links:" array >>>
                        if len(stack) == 3 and in_rail_links_array:
                            in_rail_links_array = False
                        # <<< MODIFIED: Include 'slidenodes' >>>
                        if len(stack) == 1: current_section_name = None # Exiting nodes/beams/torsionbars/slidenodes array
                        if len(stack) == 0: current_part_name = None
                        in_dict = prev_in_dict
                        pos_in_arr = prev_key_or_idx + 1 if not prev_in_dict else 0 # Restore position in parent
                    else: in_dict = None
                elif node_type == '}':
                     pass # Should not be reached if structure is valid array
                else: # Value node within array
                    pos_in_arr += 1
            i += 1
        # --- End AST Traversal ---

        # <<< DEBUG PRINT: Final result before processing >>>
        # print(f"[Highlight Debug] Final AST Result before processing: element_type={element_type}, node_ids={node_ids}, found_on_line={found_element_on_line}")

        # --- Further Processing & Highlighting ---
        if not found_element_on_line or element_type is None:
            # print("[Highlight Debug] No valid element found on line or type is None.") # DEBUG
            if needs_redraw: veh_render_dirty = True; _tag_redraw_3d_views(context)
            return False # No valid element found on the line within correct context

        # --- Determine Color/Width based on AST-determined element_type ---
        if element_type == 'beam':
            # Determine specific beam type color/width
            if jb_globals.curr_vdata and 'beams' in jb_globals.curr_vdata:
                target_id1, target_id2 = node_ids[0], node_ids[1]
                active_obj = context.active_object
                target_part_origin = None
                if active_obj and active_obj.data:
                    target_part_origin = active_obj.data.get(constants.MESH_JBEAM_PART)

                if target_part_origin:
                    found_beam_data = None
                    for beam_data in jb_globals.curr_vdata['beams']:
                        if isinstance(beam_data, dict) and beam_data.get('partOrigin') == target_part_origin:
                            b_id1 = beam_data.get('id1:')
                            b_id2 = beam_data.get('id2:')
                            if (b_id1 == target_id1 and b_id2 == target_id2) or \
                               (b_id1 == target_id2 and b_id2 == target_id1):
                                found_beam_data = beam_data; break
                    if found_beam_data:
                        beam_type_from_data = found_beam_data.get('beamType', '|NORMAL')
                        if beam_type_from_data == '|ANISOTROPIC': original_color = ui_props.anisotropic_beam_color; original_width = ui_props.anisotropic_beam_width
                        elif beam_type_from_data == '|SUPPORT': original_color = ui_props.support_beam_color; original_width = ui_props.support_beam_width
                        elif beam_type_from_data == '|HYDRO': original_color = ui_props.hydro_beam_color; original_width = ui_props.hydro_beam_width
                        elif beam_type_from_data == '|BOUNDED': original_color = ui_props.bounded_beam_color; original_width = ui_props.bounded_beam_width
                        elif beam_type_from_data == '|LBEAM': original_color = ui_props.lbeam_beam_color; original_width = ui_props.lbeam_beam_width
                        elif beam_type_from_data == '|PRESSURED': original_color = ui_props.pressured_beam_color; original_width = ui_props.pressured_beam_width
                        else: original_color = ui_props.beam_color; original_width = ui_props.beam_width # Default normal
                    else: # Beam definition found on line, but not in curr_vdata (maybe newly added?)
                        original_color = ui_props.beam_color; original_width = ui_props.beam_width # Use default normal
                else: # No target part origin found? Use default normal
                    original_color = ui_props.beam_color; original_width = ui_props.beam_width
            else: # No beams in curr_vdata? Use default normal
                 original_color = ui_props.beam_color; original_width = ui_props.beam_width

        elif element_type == 'rail':
             # print("[Highlight Debug] Element identified as RAIL by AST.") # DEBUG
             original_color = ui_props.rail_color
             original_width = ui_props.rail_width
        elif element_type == 'torsionbar':
             original_color = ui_props.torsionbar_color
             original_mid_color = ui_props.torsionbar_mid_color
             original_width = ui_props.torsionbar_width
        # <<< ADDED: Slidenode color/width (uses rail settings) >>>
        elif element_type == 'slidenode':
             original_color = ui_props.rail_color # Use rail color for the line segment
             original_width = ui_props.rail_width
        # <<< END ADDED >>>

        # --- Common Logic: Find Node Positions and Set Highlight ---
        if element_type == 'node':
            # Node highlight logic remains the same
            jb_globals.highlighted_element_type = 'node'
            jb_globals.highlighted_node_ids.update(node_ids)
            jb_globals.highlighted_element_ordered_node_ids = node_ids
            veh_render_dirty = True
            _tag_redraw_3d_views(context)
            return True

        # --- Find Node Positions (For Beams, Rails, Torsionbars, Slidenodes) ---
        active_obj = context.active_object
        active_part_name = None
        collection = None
        is_vehicle_part = False
        if active_obj and active_obj.data:
            active_part_name = active_obj.data.get(constants.MESH_JBEAM_PART)
            collection = active_obj.users_collection[0] if active_obj.users_collection else None
            is_vehicle_part = collection is not None and collection.get(constants.COLLECTION_VEHICLE_MODEL) is not None

        temp_node_map = {}
        if is_vehicle_part and collection:
            if not part_name_to_obj:
                 for obj_iter in collection.all_objects:
                    if obj_iter.data and obj_iter.data.get(constants.MESH_JBEAM_PART):
                        part_name_to_obj[obj_iter.data[constants.MESH_JBEAM_PART]] = obj_iter

            for obj_iter in collection.all_objects:
                 if obj_iter.visible_get() and obj_iter.data and obj_iter.data.get(constants.MESH_JBEAM_PART) is not None:
                    obj_iter_data = obj_iter.data
                    temp_bm = None
                    try:
                        # <<< START MODIFICATION >>>
                        if obj_iter == active_obj and active_obj.mode == 'EDIT':
                            try:
                                temp_bm = bmesh.from_edit_mesh(obj_iter_data)
                            except ValueError:
                                # Mesh not ready for edit mode access yet, skip this object for now
                                # print("[Highlight Debug] Mesh not ready for edit mode access (ValueError) during initial load/draw. Skipping highlight.", file=sys.stderr)
                                if needs_redraw: veh_render_dirty = True; _tag_redraw_3d_views(context) # Ensure redraw if highlight was previously active
                                return False # Abort highlight attempt for this cycle
                        # <<< END MODIFICATION >>>
                        else: # Object mode or not the active object
                            temp_bm = bmesh.new(); temp_bm.from_mesh(obj_iter_data)

                        node_id_layer = temp_bm.verts.layers.string.get(constants.VL_NODE_ID)
                        is_fake_layer = temp_bm.verts.layers.int.get(constants.VL_NODE_IS_FAKE)
                        if node_id_layer and is_fake_layer:
                            temp_bm.verts.ensure_lookup_table()
                            obj_matrix_copy = obj_iter.matrix_world.copy()
                            for v in temp_bm.verts:
                                if v[is_fake_layer] == 0:
                                    nid = v[node_id_layer].decode('utf-8')
                                    temp_node_map[nid] = (v.co.copy(), obj_matrix_copy)
                    finally:
                        if temp_bm and not (obj_iter == active_obj and active_obj.mode == 'EDIT'): temp_bm.free()
        elif active_obj and active_part_name: # Single part import
             temp_bm = None
             try:
                # <<< START MODIFICATION >>>
                if active_obj.mode == 'EDIT':
                    try:
                        temp_bm = bmesh.from_edit_mesh(active_obj.data)
                    except ValueError:
                        # Mesh not ready for edit mode access yet, skip highlight
                        # print("[Highlight Debug] Mesh not ready for edit mode access (ValueError) during initial load/draw. Skipping highlight.", file=sys.stderr)
                        if needs_redraw: veh_render_dirty = True; _tag_redraw_3d_views(context) # Ensure redraw if highlight was previously active
                        return False # Abort highlight attempt for this cycle
                # <<< END MODIFICATION >>>
                else: # Object mode
                    temp_bm = bmesh.new(); temp_bm.from_mesh(active_obj.data)

                node_id_layer = temp_bm.verts.layers.string.get(constants.VL_NODE_ID)
                is_fake_layer = temp_bm.verts.layers.int.get(constants.VL_NODE_IS_FAKE)
                if node_id_layer and is_fake_layer:
                    temp_bm.verts.ensure_lookup_table()
                    obj_matrix_copy = active_obj.matrix_world.copy()
                    for v in temp_bm.verts:
                        if v[is_fake_layer] == 0:
                            nid = v[node_id_layer].decode('utf-8')
                            temp_node_map[nid] = (v.co.copy(), obj_matrix_copy)
             finally:
                 if temp_bm and not (active_obj.mode == 'EDIT'): temp_bm.free()

        # --- Get World Positions and Check Origins ---
        world_positions = [] # Used for beams, rails, torsionbars
        missing_nodes = []
        node_origins = {} # Used for beams, rails, torsionbars
        highlight_set = False # <<< Initialize highlight_set >>>

        # <<< ADDED: Slidenode specific position finding >>>
        if element_type == 'slidenode':
            rail_node_id1 = None
            rail_node_id2 = None
            # Find the rail definition in curr_vdata
            if jb_globals.curr_vdata and 'rails' in jb_globals.curr_vdata:
                rails_data = jb_globals.curr_vdata['rails']
                if isinstance(rails_data, dict):
                    rail_info = rails_data.get(slidenode_rail_name)
                    if isinstance(rail_info, list) and len(rail_info) == 2:
                        rail_node_id1, rail_node_id2 = rail_info[0], rail_info[1]
                    elif isinstance(rail_info, dict):
                        links = rail_info.get('links:')
                        if isinstance(links, list) and len(links) == 2:
                            rail_node_id1, rail_node_id2 = links[0], links[1]

            if rail_node_id1 is None or rail_node_id2 is None:
                print(f"Warning: Could not find rail definition for '{slidenode_rail_name}'")
                if needs_redraw: veh_render_dirty = True; _tag_redraw_3d_views(context)
                return False # Cannot highlight if rail nodes not found

            # Now find positions for slidenode_node_id, rail_node_id1, rail_node_id2
            node_ids_to_find = [slidenode_node_id, rail_node_id1, rail_node_id2]
            slidenode_world_positions = {} # Use a dict for slidenode specific positions

            for node_id in node_ids_to_find:
                wp = None
                pos_data = temp_node_map.get(node_id)
                cache_data = all_nodes_cache.get(node_id)
                if pos_data:
                    wp = pos_data[1] @ pos_data[0]
                elif cache_data:
                    wp = cache_data[0]

                if wp is None:
                    missing_nodes.append(node_id)
                slidenode_world_positions[node_id] = wp

            if missing_nodes:
                print(f"[Highlight Debug] Missing nodes for slidenode highlight: {missing_nodes}")
                if needs_redraw: veh_render_dirty = True; _tag_redraw_3d_views(context)
                return False

            # Populate highlight data for slidenode
            jb_globals.highlighted_element_type = 'slidenode'
            jb_globals.highlighted_node_ids.add(slidenode_node_id) # Add only the slidenode's ID for text coloring
            # Add rail segment coordinates
            highlight_coords.extend([slidenode_world_positions[rail_node_id1], slidenode_world_positions[rail_node_id2]])
            jb_globals.highlighted_element_color = original_color # Use rail color set earlier
            jb_globals.highlighted_element_original_width = original_width # Use rail width set earlier
            highlight_set = True
        # <<< END ADDED: Slidenode specific position finding >>>
        else: # Logic for beams, rails, torsionbars
            for node_id in node_ids: # Use the ordered list
                wp = None
                pos_data = temp_node_map.get(node_id)
                cache_data = all_nodes_cache.get(node_id)

                if pos_data:
                    wp = pos_data[1] @ pos_data[0]
                    found_origin = None
                    if is_vehicle_part and collection:
                        for obj_iter in collection.all_objects:
                            if (obj_iter.matrix_world - pos_data[1]).is_zero(1e-5):
                                 if obj_iter.data and obj_iter.data.get(constants.MESH_JBEAM_PART):
                                     found_origin = obj_iter.data[constants.MESH_JBEAM_PART]; break
                    elif active_part_name: found_origin = active_part_name
                    node_origins[node_id] = found_origin if found_origin else '?'

                elif cache_data:
                    wp = cache_data[0]
                    node_origins[node_id] = cache_data[2]

                if wp is None: missing_nodes.append(node_id)
                world_positions.append(wp)

            if missing_nodes:
                # print(f"[Highlight Debug] Missing nodes for highlight: {missing_nodes}") # DEBUG
                if needs_redraw: veh_render_dirty = True; _tag_redraw_3d_views(context)
                return False

            # --- Populate Coordinate Lists and Finalize Highlight ---
            if element_type == 'beam':
                is_cross_part_candidate = False
                if len(node_origins) == 2:
                    origin1 = node_origins.get(node_ids[0], '?')
                    origin2 = node_origins.get(node_ids[1], '?')
                    if origin1 != origin2 and '?' not in {origin1, origin2}:
                        is_cross_part_candidate = True

                id1_in_active_geom = node_ids[0] in temp_node_map
                id2_in_active_geom = node_ids[1] in temp_node_map

                if is_cross_part_candidate and id1_in_active_geom and id2_in_active_geom:
                     # Use correct normal color based on beamType (determined earlier)
                     # The original_color/width are already set correctly for the beam type
                     highlight_coords.extend([world_positions[0], world_positions[1]])
                     highlight_set = True
                elif is_cross_part_candidate:
                    element_type = 'cross_part_beam' # Reclassify for drawing
                    original_color = ui_props.cross_part_beam_color # Use cross-part color
                    original_width = ui_props.cross_part_beam_width
                    highlight_coords.extend([world_positions[0], world_positions[1]])
                    highlight_set = True
                else: # Normal beam within the same part
                    # Use the color/width determined earlier based on beamType
                    highlight_coords.extend([world_positions[0], world_positions[1]])
                    highlight_set = True

            elif element_type == 'torsionbar':
                highlight_torsionbar_outer_coords.extend([world_positions[0], world_positions[1]])
                highlight_torsionbar_mid_coords.extend([world_positions[1], world_positions[2]])
                highlight_torsionbar_outer_coords.extend([world_positions[2], world_positions[3]])
                highlight_set = True
                # Colors/width already set

            elif element_type == 'rail':
                 # print("[Highlight Debug] Populating highlight_coords for RAIL.") # DEBUG
                 # Cross-part check for rails
                 is_cross_part_candidate = False
                 if len(node_origins) == 2:
                     origin1 = node_origins.get(node_ids[0], '?')
                     origin2 = node_origins.get(node_ids[1], '?')
                     if origin1 != origin2 and '?' not in {origin1, origin2}:
                         is_cross_part_candidate = True

                 id1_in_active_geom = node_ids[0] in temp_node_map
                 id2_in_active_geom = node_ids[1] in temp_node_map

                 if is_cross_part_candidate and id1_in_active_geom and id2_in_active_geom:
                     # Still a rail, just happens to use nodes from different parts' geometry
                     original_color = ui_props.rail_color # Use rail color
                     original_width = ui_props.rail_width
                     highlight_coords.extend([world_positions[0], world_positions[1]])
                     highlight_set = True
                 elif is_cross_part_candidate:
                      element_type = 'cross_part_beam' # Reclassify for drawing as cross-part
                      original_color = ui_props.cross_part_beam_color
                      original_width = ui_props.cross_part_beam_width
                      highlight_coords.extend([world_positions[0], world_positions[1]])
                      highlight_set = True
                 else: # Normal rail within the same part
                     # Colors/width already set
                     highlight_coords.extend([world_positions[0], world_positions[1]])
                     highlight_set = True

        # --- Finalize Highlight State ---
        if highlight_set:
            # print(f"[Highlight Debug] Highlight SET: type={element_type}, color={original_color}, width={original_width}") # DEBUG
            jb_globals.highlighted_element_type = element_type
            jb_globals.highlighted_element_color = original_color
            jb_globals.highlighted_element_mid_color = original_mid_color
            jb_globals.highlighted_element_original_width = original_width
            # Only add node IDs for non-slidenode types here, slidenode handled above
            if element_type != 'slidenode':
                jb_globals.highlighted_node_ids.update(node_ids)
                jb_globals.highlighted_element_ordered_node_ids = node_ids
            veh_render_dirty = True
            _tag_redraw_3d_views(context)
        else:
             # print("[Highlight Debug] Highlight NOT SET.") # DEBUG
             if needs_redraw: veh_render_dirty = True; _tag_redraw_3d_views(context)

        return highlight_set

    except Exception as e:
        print(f"[Highlight Debug] EXCEPTION in find_and_highlight_element_for_line: {e}", file=sys.stderr)
        traceback.print_exc()
        # Clear highlight on exception
        highlight_coords.clear()
        highlight_torsionbar_outer_coords.clear()
        highlight_torsionbar_mid_coords.clear()
        jb_globals.highlighted_node_ids.clear()
        jb_globals.highlighted_element_ordered_node_ids.clear()
        if needs_redraw: veh_render_dirty = True; _tag_redraw_3d_views(context)
        return False
# <<< END MODIFIED FUNCTION find_and_highlight_element_for_line >>>

# --- END MOVED HIGHLIGHT LOGIC ---


# Draws Node IDs and tooltips
def draw_callback_px(context: bpy.types.Context):
    global part_name_to_obj
    scene = context.scene
    ui_props = scene.ui_properties
    if not hasattr(context, 'scene') or not hasattr(scene, 'ui_properties'):
        return
    font_id = 0

    active_obj = context.active_object
    is_valid_jbeam_obj = False
    is_selected = False
    is_editing_enabled = False
    if active_obj and active_obj.data and active_obj.data.get(constants.MESH_JBEAM_PART) is not None:
        is_valid_jbeam_obj = True
        is_editing_enabled = active_obj.data.get(constants.MESH_EDITING_ENABLED, False)
        if active_obj in context.selected_objects:
            is_selected = True

    should_draw = is_valid_jbeam_obj and is_selected # Allow drawing even if editing disabled for tooltips/IDs
    if not should_draw: return
    active_obj_data = active_obj.data

    collection = active_obj.users_collection[0] if active_obj.users_collection else None
    is_vehicle_part = collection is not None and collection.get(constants.COLLECTION_VEHICLE_MODEL) is not None

    bm = None
    try:
        # Only get bmesh if editing is enabled and in edit mode
        if is_editing_enabled and active_obj.mode == 'EDIT':
            bm = bmesh.from_edit_mesh(active_obj_data)
        elif not is_vehicle_part: # For single part, get bmesh even in object mode for drawing
            bm = bmesh.new(); bm.from_mesh(active_obj_data)
    except Exception as e:
        print(f"Error accessing bmesh for {active_obj.name}: {e}", file=sys.stderr)
        if not is_vehicle_part: return

    ctxRegion = context.region
    ctxRegionData = context.region_data
    lblfPosition = blfpos; lblfDraw = blfdraw; lblfDims = blfdims
    blfsize(font_id, ui_props.node_id_font_size)
    default_color = (1.0, 1.0, 1.0, 1.0)
    selected_color = (0.0, 0.85, 0.0, 1.0) # Darker-Green for highlighted nodes (text editor)
    yellow_color = (1.0, 1.0, 0.0, 1.0) # Yellow color for viewport selection
    orange_color = (1.0, 0.5, 0.0, 1.0) # Orange color for both selected and highlighted
    black_color = (0.0, 0.0, 0.0, 1.0)
    outline_size = ui_props.node_id_outline_size
    highlighted_cross_part_color = (1.0, 0.5, 1.0, 1.0)
    # <<< ADDED: Slidenode color >>>
    slidenode_color = (1.0, 0.7, 0.7, 1.0) # Light pink

    def draw_text_with_outline(font_id, text, x, y, text_color):
        # ... (outline drawing logic remains the same) ...
        if outline_size > 0:
            blfcolor(font_id, *black_color)
            lblfPosition(font_id, x - outline_size, y, 0); lblfDraw(font_id, text)
            lblfPosition(font_id, x + outline_size, y, 0); lblfDraw(font_id, text)
            lblfPosition(font_id, x, y - outline_size, 0); lblfDraw(font_id, text)
            lblfPosition(font_id, x, y + outline_size, 0); lblfDraw(font_id, text)
            if outline_size > 1:
                 lblfPosition(font_id, x - outline_size, y - outline_size, 0); lblfDraw(font_id, text)
                 lblfPosition(font_id, x + outline_size, y - outline_size, 0); lblfDraw(font_id, text)
                 lblfPosition(font_id, x - outline_size, y + outline_size, 0); lblfDraw(font_id, text)
                 lblfPosition(font_id, x + outline_size, y + outline_size, 0); lblfDraw(font_id, text)
        blfcolor(font_id, *text_color)
        lblfPosition(font_id, x, y, 0)
        lblfDraw(font_id, text)

    # <<< START NEW LOGIC >>>
    # Set to store node IDs defined by the active object's mesh
    active_object_defined_node_ids = set()
    # <<< END NEW LOGIC >>>

    # --- Node ID Drawing ---
    if ui_props.toggle_node_ids_text:
        selected_indices_set = set()
        if is_editing_enabled and active_obj.mode == 'EDIT':
            selected_indices_set = {idx for idx, _ in jb_globals.selected_nodes}

        # Use the SET for quick lookup here
        highlighted_nodes = jb_globals.highlighted_node_ids

        if is_vehicle_part:
            part_name_to_obj.clear()
            for obj in collection.all_objects:
                if obj.data and obj.data.get(constants.MESH_JBEAM_PART):
                     part_name_to_obj[obj.data[constants.MESH_JBEAM_PART]] = obj

            for part_name, obj in part_name_to_obj.items():
                if not obj.visible_get(): continue
                part_bm = None; obj_data = obj.data
                try:
                    # Use the bmesh already obtained for the active object if applicable
                    if obj == active_obj and bm is not None:
                        part_bm = bm
                    else: # Otherwise, create a temporary one
                        part_bm = bmesh.new(); part_bm.from_mesh(obj_data)

                    node_id_layer = part_bm.verts.layers.string.get(constants.VL_NODE_ID)
                    is_fake_layer = part_bm.verts.layers.int.get(constants.VL_NODE_IS_FAKE)
                    if not node_id_layer or not is_fake_layer:
                        if part_bm != bm and part_bm: part_bm.free()
                        continue
                    part_bm.verts.ensure_lookup_table()

                    for v in part_bm.verts:
                        if v[is_fake_layer] == 1 or v.hide: continue
                        coord = obj.matrix_world @ v.co
                        node_id = v[node_id_layer].decode('utf-8')

                        # <<< START NEW LOGIC >>>
                        # If this node belongs to the currently active object, add it to the set
                        if obj == active_obj:
                            active_object_defined_node_ids.add(node_id)
                        # <<< END NEW LOGIC >>>

                        pos_text = location_3d_to_region_2d(ctxRegion, ctxRegionData, coord)
                        if pos_text:
                            # <<< MODIFIED: Color logic for slidenodes >>>
                            text_color = default_color
                            is_selected_in_viewport = obj == active_obj and is_editing_enabled and v.index in selected_indices_set
                            is_highlighted_by_text = node_id in highlighted_nodes # Use the set here
                            is_slidenode_highlight = jb_globals.highlighted_element_type == 'slidenode' and is_highlighted_by_text

                            if is_selected_in_viewport and is_slidenode_highlight:
                                text_color = orange_color # Keep orange for combined selection
                            elif is_slidenode_highlight:
                                text_color = slidenode_color # Pink for the slidenode itself
                            elif is_selected_in_viewport and is_highlighted_by_text: # Standard highlight + viewport select
                                text_color = orange_color
                            elif is_highlighted_by_text: # Standard text highlight
                                text_color = selected_color
                            elif is_selected_in_viewport: # Viewport select only
                                 text_color = yellow_color
                            # <<< END MODIFIED >>>

                            draw_text_with_outline(font_id, node_id, pos_text[0], pos_text[1], text_color)
                except Exception as e: print(f"Error processing part {obj.name} for drawing: {e}", file=sys.stderr)
                finally:
                     # Only free the temporary bmesh
                     if part_bm != bm and part_bm: part_bm.free()

        elif bm: # Single Part (bm is guaranteed to be for the active object here)
            node_id_layer = bm.verts.layers.string.get(constants.VL_NODE_ID)
            is_fake_layer = bm.verts.layers.int.get(constants.VL_NODE_IS_FAKE)
            if node_id_layer and is_fake_layer:
                bm.verts.ensure_lookup_table()
                for v in bm.verts:
                    if v[is_fake_layer] == 1 or v.hide: continue
                    coord = active_obj.matrix_world @ v.co
                    node_id = v[node_id_layer].decode('utf-8')

                    # <<< START NEW LOGIC >>>
                    # Add node ID to the set (always the active object in this case)
                    active_object_defined_node_ids.add(node_id)
                    # <<< END NEW LOGIC >>>

                    pos_text = location_3d_to_region_2d(ctxRegion, ctxRegionData, coord)
                    if pos_text:
                        # <<< MODIFIED: Color logic for slidenodes >>>
                        text_color = default_color
                        is_selected_in_viewport = is_editing_enabled and v.index in selected_indices_set
                        is_highlighted_by_text = node_id in highlighted_nodes # Use the set here
                        is_slidenode_highlight = jb_globals.highlighted_element_type == 'slidenode' and is_highlighted_by_text

                        if is_selected_in_viewport and is_slidenode_highlight:
                            text_color = orange_color # Keep orange for combined selection
                        elif is_slidenode_highlight:
                            text_color = slidenode_color # Pink for the slidenode itself
                        elif is_selected_in_viewport and is_highlighted_by_text: # Standard highlight + viewport select
                            text_color = orange_color
                        elif is_highlighted_by_text: # Standard text highlight
                            text_color = selected_color
                        elif is_selected_in_viewport: # Viewport select only
                             text_color = yellow_color
                        # <<< END MODIFIED >>>

                        draw_text_with_outline(font_id, node_id, pos_text[0], pos_text[1], text_color)

    # --- Cross-Part Node ID Drawing ---
    if ui_props.toggle_node_ids_text and all_nodes_cache:
        cross_part_color = ui_props.cross_part_beam_color
        target_other_part_node_ids = set()
        active_part_name = active_obj_data.get(constants.MESH_JBEAM_PART)
        active_filepath = active_obj_data.get(constants.MESH_JBEAM_FILE_PATH)
        highlighted_nodes = jb_globals.highlighted_node_ids # Use the set here

        # Logic to populate target_other_part_node_ids remains the same
        if active_part_name and active_filepath:
            part_data = None
            if jb_globals.curr_vdata and active_part_name in jb_globals.curr_vdata:
                part_data = jb_globals.curr_vdata[active_part_name]
            else:
                full_file_data, _ = jbeam_io.get_jbeam(active_filepath, True, False)
                if full_file_data and active_part_name in full_file_data:
                    part_data = full_file_data[active_part_name]
                # else: warnings already printed

            if isinstance(part_data, dict):
                # Check Beams
                if 'beams' in part_data and isinstance(part_data['beams'], list):
                    for beam in part_data['beams']:
                        id1, id2 = None, None
                        if isinstance(beam, dict): id1, id2 = beam.get('id1:'), beam.get('id2:')
                        elif isinstance(beam, list) and len(beam) >= 2: id1, id2 = beam[0], beam[1]
                        if id1 and id2:
                            node1_cache_data = all_nodes_cache.get(id1)
                            node2_cache_data = all_nodes_cache.get(id2)
                            if node1_cache_data and node1_cache_data[2] != active_part_name: target_other_part_node_ids.add(id1)
                            if node2_cache_data and node2_cache_data[2] != active_part_name: target_other_part_node_ids.add(id2)
                # Check Torsionbars
                if 'torsionbars' in part_data and isinstance(part_data['torsionbars'], list):
                    for tb in part_data['torsionbars']:
                        tb_node_ids = []
                        if isinstance(tb, dict): tb_node_ids = [tb.get(f'id{i}:') for i in range(1, 5)]
                        elif isinstance(tb, list) and len(tb) >= 4: tb_node_ids = tb[:4]
                        if len(tb_node_ids) == 4 and all(isinstance(nid, str) for nid in tb_node_ids):
                            for node_id in tb_node_ids:
                                cache_data = all_nodes_cache.get(node_id)
                                if cache_data and cache_data[2] != active_part_name: target_other_part_node_ids.add(node_id)
                # Check Rails
                if 'rails' in part_data and isinstance(part_data['rails'], dict):
                    for rail_name, rail_info in part_data['rails'].items():
                        rail_node_ids = None
                        if isinstance(rail_info, list) and len(rail_info) == 2: rail_node_ids = rail_info
                        elif isinstance(rail_info, dict): rail_node_ids = rail_info.get('links:')
                        if isinstance(rail_node_ids, list) and len(rail_node_ids) == 2:
                            for node_id in rail_node_ids:
                                cache_data = all_nodes_cache.get(node_id)
                                if cache_data and cache_data[2] != active_part_name: target_other_part_node_ids.add(node_id)
        # End logic to populate target_other_part_node_ids

        # Iterate through cache to draw cross-part nodes
        for node_id, (world_pos, _, part_origin) in all_nodes_cache.items():
            # <<< START NEW CHECK >>>
            # Skip drawing if this node ID is part of the active object's mesh geometry
            if node_id in active_object_defined_node_ids:
                continue
            # <<< END NEW CHECK >>>

            # Only draw if this node is targeted as an 'other part' node by the active part
            if node_id in target_other_part_node_ids:
                pos_text = location_3d_to_region_2d(ctxRegion, ctxRegionData, world_pos)
                if pos_text:
                    text_color = cross_part_color
                    if node_id in highlighted_nodes: # Check if it's also highlighted by text editor click (use set)
                        text_color = highlighted_cross_part_color # Use the brighter highlight color
                    draw_text_with_outline(font_id, node_id, pos_text[0], pos_text[1], text_color)


    # --- Tooltip Positioning & Drawing --- <<< MODIFIED >>>
    padding_x = 65; padding_y = 20
    region_width = ctxRegion.width; region_height = ctxRegion.height
    line_height = lblfDims(font_id, "X")[1]; line_padding = 4
    tooltip_placement = ui_props.tooltip_placement
    est_max_width = 250
    if tooltip_placement == 'BOTTOM_LEFT': bottom_left_x = padding_x
    elif tooltip_placement == 'BOTTOM_CENTER': bottom_left_x = region_width / 2 - est_max_width / 2
    elif tooltip_placement == 'BOTTOM_RIGHT': bottom_left_x = region_width - est_max_width - padding_x
    else: bottom_left_x = padding_x

    if is_editing_enabled:
        # --- Beam Tooltips ---
        beam_params_height = 0; beam_line_height_offset = 0
        if ui_props.toggle_params_tooltip and jb_globals._selected_beam_params_info is not None: # Use shared toggle
            params_list = jb_globals._selected_beam_params_info.get('params_list')
            if params_list: beam_params_height = len(params_list) * (line_height + line_padding)

        if ui_props.toggle_line_tooltip and jb_globals._selected_beam_line_info is not None: # Use shared toggle
            line_num = jb_globals._selected_beam_line_info.get('line')
            if line_num is not None:
                beam_line_y = padding_y + beam_params_height
                beam_line_height_offset = line_height + line_padding
                draw_text_with_outline(font_id, f"Line: {line_num}", bottom_left_x, beam_line_y, ui_props.line_tooltip_color) # Use shared color

        if ui_props.toggle_params_tooltip and jb_globals._selected_beam_params_info is not None: # Use shared toggle
            params_list = jb_globals._selected_beam_params_info.get('params_list')
            if params_list:
                name_color = ui_props.params_tooltip_color; value_color = ui_props.params_value_tooltip_color # Use shared colors
                start_y = padding_y + (len(params_list) - 1) * (line_height + line_padding)
                for i, (key, value_repr) in enumerate(params_list):
                    current_y = start_y - (i * (line_height + line_padding)); key_text = f"{key}: "
                    draw_text_with_outline(font_id, key_text, bottom_left_x, current_y, name_color)
                    key_width = lblfDims(font_id, key_text)[0]; value_x = bottom_left_x + key_width
                    draw_text_with_outline(font_id, value_repr, value_x, current_y, value_color)

        # --- Node Tooltips ---
        node_params_height = 0; node_line_height_offset = 0
        total_beam_tooltip_height = beam_params_height + beam_line_height_offset
        if ui_props.toggle_params_tooltip and jb_globals._selected_node_params_info is not None: # Use shared toggle
            params_list = jb_globals._selected_node_params_info.get('params_list')
            if params_list: node_params_height = len(params_list) * (line_height + line_padding)

        if ui_props.toggle_line_tooltip and jb_globals._selected_node_line_info is not None: # Use shared toggle
            line_num = jb_globals._selected_node_line_info.get('line')
            if line_num is not None:
                node_line_y = padding_y + total_beam_tooltip_height + node_params_height
                node_line_height_offset = line_height + line_padding
                draw_text_with_outline(font_id, f"Line: {line_num}", bottom_left_x, node_line_y, ui_props.line_tooltip_color) # Use shared color

        if ui_props.toggle_params_tooltip and jb_globals._selected_node_params_info is not None: # Use shared toggle
            params_list = jb_globals._selected_node_params_info.get('params_list')
            if params_list:
                name_color = ui_props.params_tooltip_color; value_color = ui_props.params_value_tooltip_color # Use shared colors
                start_y = padding_y + total_beam_tooltip_height + (len(params_list) - 1) * (line_height + line_padding)
                for i, (key, value_repr) in enumerate(params_list):
                    current_y = start_y - (i * (line_height + line_padding)); key_text = f"{key}: "
                    draw_text_with_outline(font_id, key_text, bottom_left_x, current_y, name_color)
                    key_width = lblfDims(font_id, key_text)[0]; value_x = bottom_left_x + key_width
                    draw_text_with_outline(font_id, value_repr, value_x, current_y, value_color)
    # --- End Tooltip Positioning & Drawing ---

    # Final cleanup
    if bm and not is_vehicle_part and active_obj.mode != 'EDIT':
        bm.free()

# Draws beams, rails, torsionbars
def draw_callback_view(context: bpy.types.Context):
    global veh_render_dirty, beam_render_shader
    global beam_render_batch, beam_coords
    global anisotropic_beam_render_batch, anisotropic_beam_coords
    global support_beam_render_batch, support_beam_coords
    global hydro_beam_render_batch, hydro_beam_coords
    global bounded_beam_render_batch, bounded_beam_coords
    global lbeam_render_batch, lbeam_coords
    global pressured_beam_render_batch, pressured_beam_coords
    global torsionbar_render_batch, torsionbar_coords
    global torsionbar_red_render_batch, torsionbar_red_coords
    global rail_render_batch, rail_coords
    global cross_part_beam_render_batch, cross_part_beam_coords
    global all_nodes_cache, all_nodes_cache_dirty
    # Highlight batches
    global highlight_render_batch, highlight_coords
    global highlight_torsionbar_outer_batch, highlight_torsionbar_outer_coords
    global highlight_torsionbar_mid_batch, highlight_torsionbar_mid_coords
    global warned_missing_nodes_this_rebuild # <<< ADD global declaration

    scene = context.scene
    ui_props = scene.ui_properties
    if not hasattr(context, 'scene') or not hasattr(scene, 'ui_properties'): return

    active_obj = context.active_object
    is_valid_jbeam_obj = False; is_selected = False
    if active_obj and active_obj.data and active_obj.data.get(constants.MESH_JBEAM_PART) is not None:
        is_valid_jbeam_obj = True
        if active_obj in context.selected_objects: is_selected = True

    should_draw = is_valid_jbeam_obj and is_selected
    if not should_draw:
        # ... (batch clearing logic remains the same) ...
        batches_were_cleared = False
        if beam_render_batch: beam_render_batch = None; batches_were_cleared = True
        if anisotropic_beam_render_batch: anisotropic_beam_render_batch = None; batches_were_cleared = True
        if support_beam_render_batch: support_beam_render_batch = None; batches_were_cleared = True
        if hydro_beam_render_batch: hydro_beam_render_batch = None; batches_were_cleared = True
        if bounded_beam_render_batch: bounded_beam_render_batch = None; batches_were_cleared = True
        if lbeam_render_batch: lbeam_render_batch = None; batches_were_cleared = True
        if pressured_beam_render_batch: pressured_beam_render_batch = None; batches_were_cleared = True
        if torsionbar_render_batch: torsionbar_render_batch = None; batches_were_cleared = True
        if torsionbar_red_render_batch: torsionbar_red_render_batch = None; batches_were_cleared = True
        if rail_render_batch: rail_render_batch = None; batches_were_cleared = True
        if cross_part_beam_render_batch: cross_part_beam_render_batch = None; batches_were_cleared = True
        # Clear highlight batches
        if highlight_render_batch: highlight_render_batch = None; batches_were_cleared = True
        if highlight_torsionbar_outer_batch: highlight_torsionbar_outer_batch = None; batches_were_cleared = True
        if highlight_torsionbar_mid_batch: highlight_torsionbar_mid_batch = None; batches_were_cleared = True
        if batches_were_cleared:
            beam_coords.clear(); anisotropic_beam_coords.clear(); support_beam_coords.clear()
            hydro_beam_coords.clear(); bounded_beam_coords.clear(); lbeam_coords.clear()
            pressured_beam_coords.clear(); torsionbar_coords.clear(); torsionbar_red_coords.clear()
            rail_coords.clear(); cross_part_beam_coords.clear()
            # Clear highlight coords
            highlight_coords.clear()
            highlight_torsionbar_outer_coords.clear()
            highlight_torsionbar_mid_coords.clear()
            veh_render_dirty = True # Use local drawing global
        return

    if beam_render_shader is None:
        beam_render_shader = gpu.shader.from_builtin('UNIFORM_COLOR')

    # Use scene property to check if dirty flag was set by property updates
    if scene.jbeam_editor_veh_render_dirty:
        veh_render_dirty = True
        scene.jbeam_editor_veh_render_dirty = False # Reset scene flag

    # Check if node cache needs updating (always check if dirty)
    if all_nodes_cache_dirty:
        update_all_nodes_cache(context)
        veh_render_dirty = True # Force batch rebuild after cache update

    # <<< MODIFIED: Check if batches need rebuilding, considering coordinate lists >>>
    batches_missing = (
        (ui_props.toggle_beams_vis and beam_render_batch is None and beam_coords) or # Add coord check
        (ui_props.toggle_anisotropic_beams_vis and anisotropic_beam_render_batch is None and anisotropic_beam_coords) or # Add coord check
        (ui_props.toggle_support_beams_vis and support_beam_render_batch is None and support_beam_coords) or # Add coord check
        (ui_props.toggle_hydro_beams_vis and hydro_beam_render_batch is None and hydro_beam_coords) or # Add coord check
        (ui_props.toggle_bounded_beams_vis and bounded_beam_render_batch is None and bounded_beam_coords) or # Add coord check
        (ui_props.toggle_lbeam_beams_vis and lbeam_render_batch is None and lbeam_coords) or # Add coord check
        (ui_props.toggle_pressured_beams_vis and pressured_beam_render_batch is None and pressured_beam_coords) or # Add coord check
        (ui_props.toggle_torsionbars_vis and torsionbar_render_batch is None and torsionbar_coords) or # Add coord check (outer)
        (ui_props.toggle_torsionbars_vis and torsionbar_red_render_batch is None and torsionbar_red_coords) or # Add coord check (mid)
        (ui_props.toggle_rails_vis and rail_render_batch is None and rail_coords) or # Add coord check
        (ui_props.toggle_cross_part_beams_vis and cross_part_beam_render_batch is None and cross_part_beam_coords and all_nodes_cache) or # Add coord check
        # Highlight batches check remains the same as it already checks coords
        (jb_globals.highlighted_element_type == 'torsionbar' and (highlight_torsionbar_outer_batch is None or highlight_torsionbar_mid_batch is None) and (highlight_torsionbar_outer_coords or highlight_torsionbar_mid_coords)) or
        # <<< MODIFIED: Exclude 'node', include 'slidenode' >>>
        (jb_globals.highlighted_element_type not in (None, 'node') and highlight_render_batch is None and highlight_coords)
    )
    if batches_missing:
        # print("DEBUG: batches_missing = True, forcing rebuild") # Optional debug
        veh_render_dirty = True
    # <<< END MODIFIED CHECK >>>

    if veh_render_dirty:
        # print(f"DEBUG: Rebuilding batches... Time: {time.time()}") # Optional debug
        # Set to track nodes warned about *in this rebuild cycle*
        # warned_missing_nodes_this_rebuild = set() # <<< REMOVED: Now a module-level variable

        # Clear regular visualization coords
        beam_coords.clear(); anisotropic_beam_coords.clear(); support_beam_coords.clear()
        hydro_beam_coords.clear(); bounded_beam_coords.clear(); lbeam_coords.clear()
        pressured_beam_coords.clear(); torsionbar_coords.clear(); torsionbar_red_coords.clear()
        rail_coords.clear(); cross_part_beam_coords.clear()
        # <<< ADDED: Clear highlight coords >>>
        # Highlight coords are now cleared and repopulated within find_and_highlight_element_for_line
        # and the highlight repopulation logic below. No need to clear them here.

        # ... (rest of data gathering: active_obj_data, collection, node maps, etc. remains the same) ...
        active_obj_data = active_obj.data
        collection = active_obj.users_collection[0] if active_obj.users_collection else None
        is_vehicle_part = collection is not None and collection.get(constants.COLLECTION_VEHICLE_MODEL) is not None

        node_id_to_hide_status: dict[str, bool] = {}
        node_id_to_pos_matrix_map: dict[str, tuple[Vector, Matrix]] = {}
        current_part_name = active_obj_data.get(constants.MESH_JBEAM_PART)
        edge_idx_to_beam_type_map: dict[tuple[str, int], str] = {}

        # --- Build edge_idx_to_beam_type_map (remains the same) ---
        if jb_globals.curr_vdata and 'beams' in jb_globals.curr_vdata:
            beam_part_counters = {}
            for global_beam_idx, beam_data in enumerate(jb_globals.curr_vdata['beams']):
                # Check if beam_data is a dict before accessing
                if isinstance(beam_data, dict):
                    part_origin = beam_data.get('partOrigin')
                    beam_type = beam_data.get('beamType', '|NORMAL')
                    if part_origin:
                        current_idx_in_part = beam_part_counters.get(part_origin, 0) + 1
                        beam_part_counters[part_origin] = current_idx_in_part
                        edge_idx_to_beam_type_map[(part_origin, current_idx_in_part)] = beam_type

        # --- Build node_id_to_hide_status and node_id_to_pos_matrix_map (remains the same) ---
        if is_vehicle_part:
            if not part_name_to_obj:
                 for obj_iter in collection.all_objects:
                    if obj_iter.data and obj_iter.data.get(constants.MESH_JBEAM_PART):
                        part_name_to_obj[obj_iter.data[constants.MESH_JBEAM_PART]] = obj_iter

            for obj_iter in collection.all_objects:
                if obj_iter.visible_get() and obj_iter.data and obj_iter.data.get(constants.MESH_JBEAM_PART) is not None:
                    obj_iter_data = obj_iter.data; part_name = obj_iter_data.get(constants.MESH_JBEAM_PART)
                    bm = None
                    try:
                        if obj_iter == active_obj and active_obj.mode == 'EDIT': bm = bmesh.from_edit_mesh(obj_iter_data)
                        else: bm = bmesh.new(); bm.from_mesh(obj_iter_data)

                        # ... (beam drawing logic remains the same) ...
                        beam_indices_layer = bm.edges.layers.string.get(constants.EL_BEAM_INDICES)
                        node_id_layer = bm.verts.layers.string.get(constants.VL_NODE_ID)
                        is_fake_layer = bm.verts.layers.int.get(constants.VL_NODE_IS_FAKE)
                        beam_part_origin_layer = bm.edges.layers.string.get(constants.EL_BEAM_PART_ORIGIN)

                        if node_id_layer and is_fake_layer:
                            bm.verts.ensure_lookup_table()
                            obj_matrix_copy = obj_iter.matrix_world.copy()
                            for v in bm.verts:
                                if v[is_fake_layer] == 0:
                                    node_id = v[node_id_layer].decode('utf-8')
                                    node_id_to_hide_status[node_id] = v.hide
                                    node_id_to_pos_matrix_map[node_id] = (v.co.copy(), obj_matrix_copy)

                        if beam_indices_layer and beam_part_origin_layer:
                            bm.edges.ensure_lookup_table()
                            for e in bm.edges:
                                if e.hide or any(v.hide for v in e.verts): continue
                                beam_idx_str = e[beam_indices_layer].decode('utf-8')
                                if beam_idx_str != '' and beam_idx_str != '-1':
                                    try:
                                        first_beam_idx_in_part = int(beam_idx_str.split(',')[0])
                                        edge_part_origin = e[beam_part_origin_layer].decode('utf-8')
                                        beam_type = edge_idx_to_beam_type_map.get((edge_part_origin, first_beam_idx_in_part), '|NORMAL')
                                        v1, v2 = e.verts[0], e.verts[1]
                                        world_pos1 = obj_iter.matrix_world @ v1.co; world_pos2 = obj_iter.matrix_world @ v2.co

                                        if beam_type == '|ANISOTROPIC': anisotropic_beam_coords.extend([world_pos1, world_pos2])
                                        elif beam_type == '|SUPPORT': support_beam_coords.extend([world_pos1, world_pos2])
                                        elif beam_type == '|HYDRO': hydro_beam_coords.extend([world_pos1, world_pos2])
                                        elif beam_type == '|BOUNDED': bounded_beam_coords.extend([world_pos1, world_pos2])
                                        elif beam_type == '|LBEAM': lbeam_coords.extend([world_pos1, world_pos2])
                                        elif beam_type == '|PRESSURED': pressured_beam_coords.extend([world_pos1, world_pos2])
                                        else: beam_coords.extend([world_pos1, world_pos2])
                                    except (ValueError, IndexError) as parse_err:
                                        print(f"Warning: Could not parse beam index '{beam_idx_str}' for edge in part '{part_name}'. Error: {parse_err}", file=sys.stderr)
                    except Exception as e: print(f"Error getting geometry data from {obj_iter.name}: {e}", file=sys.stderr)
                    finally:
                        if bm and not (obj_iter == active_obj and active_obj.mode == 'EDIT'): bm.free()

            # --- Torsionbar Drawing (Vehicle) --- #
            if jb_globals.curr_vdata and 'torsionbars' in jb_globals.curr_vdata and isinstance(jb_globals.curr_vdata['torsionbars'], list):
                for tb in jb_globals.curr_vdata['torsionbars']:
                    ids = []
                    # Check if tb is dict or list
                    if isinstance(tb, dict):
                        ids = [tb.get(f'id{i}:') for i in range(1, 5)]
                    elif isinstance(tb, list) and len(tb) >= 4:
                        ids = tb[:4]

                    if len(ids) != 4 or not all(ids): continue
                    if any(node_id_to_hide_status.get(id, False) for id in ids): continue

                    world_pos = [None] * 4; node_origins = [None] * 4 # <<< Store origins individually
                    all_nodes_found = True; missing_nodes = []

                    for i, node_id in enumerate(ids):
                        pos_data = node_id_to_pos_matrix_map.get(node_id)
                        cache_data = all_nodes_cache.get(node_id) # Use cache regardless of toggle
                        wp = None
                        origin = '?'

                        if pos_data:
                            wp = pos_data[1] @ pos_data[0]
                            for obj_iter in collection.all_objects:
                                if (obj_iter.matrix_world - pos_data[1]).is_zero(1e-5):
                                    if obj_iter.data and obj_iter.data.get(constants.MESH_JBEAM_PART):
                                        origin = obj_iter.data[constants.MESH_JBEAM_PART]
                                        break
                        elif cache_data: # Use cache if not in visible objects
                            wp = cache_data[0]
                            origin = cache_data[2]

                        if wp is None:
                            all_nodes_found = False
                            missing_nodes.append(node_id)

                        world_pos[i] = wp
                        node_origins[i] = origin # <<< Store origin for this node index

                    # <<< WARNING LOGIC (Use module-level set) >>>
                    if not all_nodes_found and ui_props.toggle_torsionbars_vis:
                        if any(node_id not in warned_missing_nodes_this_rebuild for node_id in missing_nodes):
                            print(f"Warning: Could not find position data for torsionbar nodes {missing_nodes}", file=sys.stderr)
                            warned_missing_nodes_this_rebuild.update(missing_nodes)
                        continue
                    # <<< END WARNING LOGIC >>>
                    # ... (Add coords if found) ...
                    torsionbar_coords.extend([world_pos[0], world_pos[1]])
                    torsionbar_red_coords.extend([world_pos[1], world_pos[2]])
                    torsionbar_coords.extend([world_pos[2], world_pos[3]])

            # --- Rail Drawing (Vehicle) ---
            if jb_globals.curr_vdata and 'rails' in jb_globals.curr_vdata and isinstance(jb_globals.curr_vdata['rails'], dict):
                 for rail_name, rail_info in jb_globals.curr_vdata['rails'].items():
                    rail_nodes = None
                    if isinstance(rail_info, list) and len(rail_info) == 2:
                        rail_nodes = rail_info
                    elif isinstance(rail_info, dict):
                        rail_nodes = rail_info.get('links:')

                    if isinstance(rail_nodes, list) and len(rail_nodes) == 2:
                        ids = rail_nodes
                        if not all(ids): continue
                        if any(node_id_to_hide_status.get(id, False) for id in ids): continue

                        world_pos = [None] * 2; node_origins = set() # Rails only have one segment, set is fine
                        all_nodes_found = True; missing_nodes = []

                        for i, node_id in enumerate(ids):
                            pos_data = node_id_to_pos_matrix_map.get(node_id)
                            cache_data = all_nodes_cache.get(node_id) # Use cache regardless of toggle
                            wp = None; origin = '?'
                            if pos_data:
                                wp = pos_data[1] @ pos_data[0]
                                for obj_iter in collection.all_objects:
                                    if (obj_iter.matrix_world - pos_data[1]).is_zero(1e-5):
                                        if obj_iter.data and obj_iter.data.get(constants.MESH_JBEAM_PART):
                                            origin = obj_iter.data[constants.MESH_JBEAM_PART]; break
                            elif cache_data: # Use cache if not in visible objects
                                wp = cache_data[0]; origin = cache_data[2]

                            if wp is None: all_nodes_found = False; missing_nodes.append(node_id)
                            else: node_origins.add(origin)
                            world_pos[i] = wp

                        # <<< WARNING LOGIC (Use module-level set) >>>
                        if not all_nodes_found and ui_props.toggle_rails_vis:
                            if any(node_id not in warned_missing_nodes_this_rebuild for node_id in missing_nodes):
                                print(f"Warning: Could not find position data for rail nodes {missing_nodes}", file=sys.stderr)
                                warned_missing_nodes_this_rebuild.update(missing_nodes)
                            continue
                        # <<< END WARNING LOGIC >>>
                        # ... (Add coords if found) ...
                        rail_coords.extend(world_pos)

            # --- Cross-Part Beam Drawing (Vehicle) ---
            # Only populate cross_part_beam_coords if the visualization toggle is ON
            if ui_props.toggle_cross_part_beams_vis and jb_globals.curr_vdata and 'beams' in jb_globals.curr_vdata:
                for beam in jb_globals.curr_vdata['beams']:
                    # Check if beam is dict before accessing
                    if not isinstance(beam, dict) or beam.get('partOrigin') != current_part_name: continue
                    id1, id2 = beam.get('id1:'), beam.get('id2:')
                    if not id1 or not id2: continue

                    # Check if both nodes exist in the currently loaded geometry for the vehicle/part
                    id1_in_active_geom = id1 in node_id_to_pos_matrix_map
                    id2_in_active_geom = id2 in node_id_to_pos_matrix_map

                    # If BOTH nodes are part of the active geometry, it's NOT a cross-part beam for drawing purposes
                    if id1_in_active_geom and id2_in_active_geom:
                        continue # Skip adding to cross_part_beam_coords

                    # Check hide status using the map (only relevant if we proceed)
                    if node_id_to_hide_status.get(id1, False) or node_id_to_hide_status.get(id2, False):
                        continue

                    # Now, determine if it *is* actually cross-part based on cache,
                    # since we know at least one node is NOT in the active geometry map.
                    cache1_data = all_nodes_cache.get(id1)
                    cache2_data = all_nodes_cache.get(id2)
                    world_pos1, world_pos2 = None, None
                    is_truly_cross_part = False
                    missing_nodes = [] # <<< ADDED: Track missing nodes for warning

                    origin1 = cache1_data[2] if cache1_data else '?'
                    origin2 = cache2_data[2] if cache2_data else '?'

                    # It's cross-part if origins differ OR if one node is missing from cache entirely (implies external)
                    # Check if origins differ AND neither is unknown ('?')
                    if origin1 != origin2 and '?' not in {origin1, origin2}:
                         is_truly_cross_part = True
                    # Also consider it cross-part if one node is in active geom and the other isn't (and not hidden)
                    elif (id1_in_active_geom and not id2_in_active_geom) or \
                         (not id1_in_active_geom and id2_in_active_geom):
                         is_truly_cross_part = True


                    if is_truly_cross_part:
                        # Get positions (prioritize active geom map, then cache)
                        if id1_in_active_geom:
                            pos1_data = node_id_to_pos_matrix_map[id1]
                            world_pos1 = pos1_data[1] @ pos1_data[0]
                        elif cache1_data:
                            world_pos1 = cache1_data[0]
                        else: # <<< ADDED: Track missing node 1
                            missing_nodes.append(id1)

                        if id2_in_active_geom:
                            pos2_data = node_id_to_pos_matrix_map[id2]
                            world_pos2 = pos2_data[1] @ pos2_data[0]
                        elif cache2_data:
                            world_pos2 = cache2_data[0]
                        else: # <<< ADDED: Track missing node 2
                            missing_nodes.append(id2)

                        # <<< WARNING LOGIC (Use module-level set) >>>
                        if missing_nodes:
                            if any(node_id not in warned_missing_nodes_this_rebuild for node_id in missing_nodes):
                                print(f"Warning: Could not find position data for beam nodes {missing_nodes}", file=sys.stderr)
                                warned_missing_nodes_this_rebuild.update(missing_nodes)
                            continue # Skip adding coords if nodes missing
                        # <<< END WARNING LOGIC >>>

                        # Add to coords only if both positions found (already checked by warning logic)
                        cross_part_beam_coords.extend([world_pos1, world_pos2])

        else: # Single Part
            if active_obj.visible_get():
                part_name = active_obj_data.get(constants.MESH_JBEAM_PART)
                bm = None
                try:
                    if active_obj.mode == 'EDIT': bm = bmesh.from_edit_mesh(active_obj_data)
                    else: bm = bmesh.new(); bm.from_mesh(active_obj_data)

                    # --- Build node maps for single part (remains the same) ---
                    node_id_layer = bm.verts.layers.string.get(constants.VL_NODE_ID)
                    is_fake_layer = bm.verts.layers.int.get(constants.VL_NODE_IS_FAKE)
                    beam_indices_layer = bm.edges.layers.string.get(constants.EL_BEAM_INDICES)
                    beam_part_origin_layer = bm.edges.layers.string.get(constants.EL_BEAM_PART_ORIGIN)

                    if node_id_layer and is_fake_layer:
                        bm.verts.ensure_lookup_table()
                        obj_matrix_copy = active_obj.matrix_world.copy()
                        for v in bm.verts:
                            if v[is_fake_layer] == 0:
                                node_id = v[node_id_layer].decode('utf-8')
                                node_id_to_hide_status[node_id] = v.hide
                                node_id_to_pos_matrix_map[node_id] = (v.co.copy(), obj_matrix_copy)

                    # --- Beam Drawing (Single Part - remains the same) ---
                    if beam_indices_layer and beam_part_origin_layer:
                        bm.edges.ensure_lookup_table()
                        for e in bm.edges:
                            if e.hide or any(v.hide for v in e.verts): continue
                            beam_idx_str = e[beam_indices_layer].decode('utf-8')
                            if beam_idx_str != '' and beam_idx_str != '-1':
                                try:
                                    first_beam_idx_in_part = int(beam_idx_str.split(',')[0])
                                    edge_part_origin = e[beam_part_origin_layer].decode('utf-8')
                                    beam_type = edge_idx_to_beam_type_map.get((edge_part_origin, first_beam_idx_in_part), '|NORMAL')
                                    v1, v2 = e.verts[0], e.verts[1]
                                    world_pos1 = active_obj.matrix_world @ v1.co; world_pos2 = active_obj.matrix_world @ v2.co

                                    if beam_type == '|ANISOTROPIC': anisotropic_beam_coords.extend([world_pos1, world_pos2])
                                    elif beam_type == '|SUPPORT': support_beam_coords.extend([world_pos1, world_pos2])
                                    elif beam_type == '|HYDRO': hydro_beam_coords.extend([world_pos1, world_pos2])
                                    elif beam_type == '|BOUNDED': bounded_beam_coords.extend([world_pos1, world_pos2])
                                    elif beam_type == '|LBEAM': lbeam_coords.extend([world_pos1, world_pos2])
                                    elif beam_type == '|PRESSURED': pressured_beam_coords.extend([world_pos1, world_pos2])
                                    else: beam_coords.extend([world_pos1, world_pos2])
                                except (ValueError, IndexError) as parse_err:
                                    print(f"Warning: Could not parse beam index '{beam_idx_str}' for edge in part '{part_name}'. Error: {parse_err}", file=sys.stderr)

                    # --- Torsionbar Drawing (Single Part) --- #
                    if jb_globals.curr_vdata and 'torsionbars' in jb_globals.curr_vdata and isinstance(jb_globals.curr_vdata['torsionbars'], list):
                        for tb in jb_globals.curr_vdata['torsionbars']:
                            ids = []
                            # Check if tb is dict or list
                            if isinstance(tb, dict):
                                ids = [tb.get(f'id{i}:') for i in range(1, 5)]
                            elif isinstance(tb, list) and len(tb) >= 4:
                                ids = tb[:4]

                            if len(ids) != 4 or not all(ids): continue
                            if any(node_id_to_hide_status.get(id, False) for id in ids): continue

                            world_pos = [None] * 4; node_origins = [None] * 4 # <<< Store origins individually
                            all_nodes_found = True; missing_nodes = []

                            for i, node_id in enumerate(ids):
                                pos_data = node_id_to_pos_matrix_map.get(node_id)
                                cache_data = all_nodes_cache.get(node_id) # Use cache regardless of toggle
                                wp = None
                                origin = '?'

                                if pos_data:
                                    wp = pos_data[1] @ pos_data[0]
                                    origin = part_name # Origin is the current part
                                elif cache_data: # Use cache if not in visible objects
                                    wp = cache_data[0]
                                    origin = cache_data[2]

                                if wp is None:
                                    all_nodes_found = False
                                    missing_nodes.append(node_id)

                                world_pos[i] = wp
                                node_origins[i] = origin # <<< Store origin for this node index

                            # <<< WARNING LOGIC (Use module-level set) >>>
                            if not all_nodes_found and ui_props.toggle_torsionbars_vis:
                                if any(node_id not in warned_missing_nodes_this_rebuild for node_id in missing_nodes):
                                    print(f"Warning: Could not find position data for torsionbar nodes {missing_nodes}", file=sys.stderr)
                                    warned_missing_nodes_this_rebuild.update(missing_nodes)
                                continue
                            # <<< END WARNING LOGIC >>>
                            # ... (Add coords if found) ...
                            torsionbar_coords.extend([world_pos[0], world_pos[1]])
                            torsionbar_red_coords.extend([world_pos[1], world_pos[2]])
                            torsionbar_coords.extend([world_pos[2], world_pos[3]])

                    # --- Rail Drawing (Single Part) ---
                    if jb_globals.curr_vdata and 'rails' in jb_globals.curr_vdata and isinstance(jb_globals.curr_vdata['rails'], dict):
                        for rail_name, rail_info in jb_globals.curr_vdata['rails'].items():
                            rail_nodes = None
                            if isinstance(rail_info, list) and len(rail_info) == 2:
                                rail_nodes = rail_info
                            elif isinstance(rail_info, dict):
                                rail_nodes = rail_info.get('links:')

                            if isinstance(rail_nodes, list) and len(rail_nodes) == 2:
                                ids = rail_nodes
                                if not all(ids): continue
                                if any(node_id_to_hide_status.get(id, False) for id in ids): continue

                                world_pos = [None] * 2; node_origins = set() # Rails only have one segment, set is fine
                                all_nodes_found = True; missing_nodes = []

                                for i, node_id in enumerate(ids):
                                    pos_data = node_id_to_pos_matrix_map.get(node_id)
                                    cache_data = all_nodes_cache.get(node_id) # Use cache regardless of toggle
                                    wp = None; origin = '?'
                                    if pos_data:
                                        wp = pos_data[1] @ pos_data[0]; origin = part_name
                                    elif cache_data: # Use cache if not in visible objects
                                        wp = cache_data[0]; origin = cache_data[2]

                                    if wp is None: all_nodes_found = False; missing_nodes.append(node_id)
                                    else: node_origins.add(origin)
                                    world_pos[i] = wp

                                # <<< WARNING LOGIC (Use module-level set) >>>
                                if not all_nodes_found and ui_props.toggle_rails_vis:
                                    if any(node_id not in warned_missing_nodes_this_rebuild for node_id in missing_nodes):
                                        print(f"Warning: Could not find position data for rail nodes {missing_nodes}", file=sys.stderr)
                                        warned_missing_nodes_this_rebuild.update(missing_nodes)
                                    continue
                                # <<< END WARNING LOGIC >>>
                                # ... (Add coords if found) ...
                                rail_coords.extend(world_pos)

                    # --- Cross-Part Beam Drawing (Single Part) ---
                    # Only populate cross_part_beam_coords if the visualization toggle is ON
                    if ui_props.toggle_cross_part_beams_vis and jb_globals.curr_vdata and 'beams' in jb_globals.curr_vdata:
                        obj_matrix = active_obj.matrix_world
                        for beam in jb_globals.curr_vdata['beams']:
                            # Check if beam is dict before accessing
                            if not isinstance(beam, dict) or beam.get('partOrigin') != current_part_name: continue
                            id1, id2 = beam.get('id1:'), beam.get('id2:')
                            if not id1 or not id2: continue

                            # Check if both nodes exist in the single part's geometry map
                            id1_in_active_geom = id1 in node_id_to_pos_matrix_map
                            id2_in_active_geom = id2 in node_id_to_pos_matrix_map

                            if id1_in_active_geom and id2_in_active_geom:
                                continue # Skip if both nodes are in the active part's geometry

                            # Check hide status
                            if node_id_to_hide_status.get(id1, False) or node_id_to_hide_status.get(id2, False):
                                continue

                            # Determine if it's truly cross-part based on cache
                            cache1_data = all_nodes_cache.get(id1)
                            cache2_data = all_nodes_cache.get(id2)
                            world_pos1, world_pos2 = None, None
                            is_truly_cross_part = False
                            missing_nodes = [] # <<< ADDED: Track missing nodes

                            origin1 = cache1_data[2] if cache1_data else '?'
                            origin2 = cache2_data[2] if cache2_data else '?'

                            # Check if origins differ AND neither is unknown ('?')
                            if origin1 != origin2 and '?' not in {origin1, origin2}:
                                is_truly_cross_part = True
                            elif (id1_in_active_geom and not id2_in_active_geom) or \
                                 (not id1_in_active_geom and id2_in_active_geom):
                                is_truly_cross_part = True

                            if is_truly_cross_part:
                                if id1_in_active_geom:
                                    pos1_data = node_id_to_pos_matrix_map[id1]
                                    world_pos1 = obj_matrix @ pos1_data[0]
                                elif cache1_data:
                                    world_pos1 = cache1_data[0]
                                else: # <<< ADDED: Track missing node 1
                                    missing_nodes.append(id1)

                                if id2_in_active_geom:
                                    pos2_data = node_id_to_pos_matrix_map[id2]
                                    world_pos2 = obj_matrix @ pos2_data[0]
                                elif cache2_data:
                                    world_pos2 = cache2_data[0]
                                else: # <<< ADDED: Track missing node 2
                                    missing_nodes.append(id2)

                                # <<< WARNING LOGIC (Use module-level set) >>>
                                if missing_nodes:
                                    if any(node_id not in warned_missing_nodes_this_rebuild for node_id in missing_nodes):
                                        print(f"Warning: Could not find position data for beam nodes {missing_nodes}", file=sys.stderr)
                                        warned_missing_nodes_this_rebuild.update(missing_nodes)
                                    continue # Skip adding coords if nodes missing
                                # <<< END WARNING LOGIC >>>

                                cross_part_beam_coords.extend([world_pos1, world_pos2])
                except Exception as e: print(f"Error getting geometry data from {active_obj.name}: {e}", file=sys.stderr)
                finally:
                    if bm and not (active_obj.mode == 'EDIT'): bm.free()

        # <<< START: Re-populate highlight coordinates >>>
        # This logic runs AFTER the main visualization coords are populated
        # It uses the highlight state set by find_and_highlight_element_for_line
        if jb_globals.highlighted_element_type is not None and jb_globals.highlighted_element_type != 'node':
            # Use the ORDERED list of node IDs for beams, rails, torsionbars
            # For slidenodes, the rail coords are already in highlight_coords
            ordered_highlight_node_ids = jb_globals.highlighted_element_ordered_node_ids
            highlight_world_positions = [] # This will be ordered correctly now
            all_highlight_nodes_found = True
            missing_highlight_nodes = []

            # Only find positions if it's not a slidenode (slidenode coords handled in find_and_highlight)
            if jb_globals.highlighted_element_type != 'slidenode':
                for node_id in ordered_highlight_node_ids:
                    wp = None
                    pos_data = node_id_to_pos_matrix_map.get(node_id)
                    cache_data = all_nodes_cache.get(node_id)

                    if pos_data:
                        wp = pos_data[1] @ pos_data[0]
                    elif cache_data:
                        wp = cache_data[0]

                    if wp is None:
                        all_highlight_nodes_found = False
                        missing_highlight_nodes.append(node_id)

                    highlight_world_positions.append(wp) # Appends in the correct order

                if not all_highlight_nodes_found:
                    if any(node_id not in warned_missing_nodes_this_rebuild for node_id in missing_highlight_nodes):
                        print(f"Warning: Could not find position data for highlighted nodes {missing_highlight_nodes}", file=sys.stderr)
                        warned_missing_nodes_this_rebuild.update(missing_highlight_nodes)
                    # Clear highlight type if nodes are missing to prevent drawing stale highlight
                    jb_globals.highlighted_element_type = None
                    jb_globals.highlighted_node_ids.clear() # Also clear the set
                    jb_globals.highlighted_element_ordered_node_ids.clear() # Also clear the list
                    highlight_coords.clear() # Clear any potentially partially populated coords
                    highlight_torsionbar_outer_coords.clear()
                    highlight_torsionbar_mid_coords.clear()
                else:
                    # Populate coordinate lists based on type (excluding slidenode)
                    element_type = jb_globals.highlighted_element_type
                    if element_type in ('beam', 'rail', 'cross_part_beam'):
                        if len(highlight_world_positions) >= 2:
                            highlight_coords.extend([highlight_world_positions[0], highlight_world_positions[1]])
                    elif element_type == 'torsionbar':
                        if len(highlight_world_positions) >= 4:
                            highlight_torsionbar_outer_coords.extend([highlight_world_positions[0], highlight_world_positions[1]])
                            highlight_torsionbar_mid_coords.extend([highlight_world_positions[1], highlight_world_positions[2]])
                            highlight_torsionbar_outer_coords.extend([highlight_world_positions[2], highlight_world_positions[3]])
            # Else: If it's a slidenode, highlight_coords should already be populated by find_and_highlight_element_for_line
            # No need to do anything here for slidenode coords.

        # <<< END: Re-populate highlight coordinates >>>


        # --- Create Batches ---
        # If coords list is empty, set batch to None. Otherwise create it.
        if beam_coords: beam_render_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": beam_coords})
        else: beam_render_batch = None
        if anisotropic_beam_coords: anisotropic_beam_render_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": anisotropic_beam_coords})
        else: anisotropic_beam_render_batch = None
        if support_beam_coords: support_beam_render_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": support_beam_coords})
        else: support_beam_render_batch = None
        if hydro_beam_coords: hydro_beam_render_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": hydro_beam_coords})
        else: hydro_beam_render_batch = None
        if bounded_beam_coords: bounded_beam_render_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": bounded_beam_coords})
        else: bounded_beam_render_batch = None
        if lbeam_coords: lbeam_render_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": lbeam_coords})
        else: lbeam_render_batch = None
        if pressured_beam_coords: pressured_beam_render_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": pressured_beam_coords})
        else: pressured_beam_render_batch = None
        if torsionbar_coords: torsionbar_render_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": torsionbar_coords})
        else: torsionbar_render_batch = None
        if torsionbar_red_coords: torsionbar_red_render_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": torsionbar_red_coords})
        else: torsionbar_red_render_batch = None
        if rail_coords: rail_render_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": rail_coords})
        else: rail_render_batch = None # Set to None if no coords
        if cross_part_beam_coords: cross_part_beam_render_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": cross_part_beam_coords})
        else: cross_part_beam_render_batch = None

        # Create highlight batches (logic remains the same, uses updated coords)
        highlight_render_batch = None
        highlight_torsionbar_outer_batch = None
        highlight_torsionbar_mid_batch = None
        if highlight_coords:
            highlight_render_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": highlight_coords})
        if highlight_torsionbar_outer_coords:
            highlight_torsionbar_outer_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": highlight_torsionbar_outer_coords})
        if highlight_torsionbar_mid_coords:
            highlight_torsionbar_mid_batch = batch_for_shader(beam_render_shader, 'LINES', {"pos": highlight_torsionbar_mid_coords})

        veh_render_dirty = False # Reset flag AFTER batch creation/clearing

    # --- Drawing ---
    # ... (GPU drawing calls remain the same) ...
    gpu.state.depth_test_set('LESS_EQUAL')

    if beam_render_batch is not None and ui_props.toggle_beams_vis:
        beam_render_shader.uniform_float("color", ui_props.beam_color); gpu.state.line_width_set(ui_props.beam_width)
        gpu.state.depth_mask_set(True); beam_render_batch.draw(beam_render_shader); gpu.state.depth_mask_set(False)

    if anisotropic_beam_render_batch is not None and ui_props.toggle_anisotropic_beams_vis:
        beam_render_shader.uniform_float("color", ui_props.anisotropic_beam_color); gpu.state.line_width_set(ui_props.anisotropic_beam_width)
        gpu.state.depth_mask_set(True); anisotropic_beam_render_batch.draw(beam_render_shader); gpu.state.depth_mask_set(False)

    if support_beam_render_batch is not None and ui_props.toggle_support_beams_vis:
        beam_render_shader.uniform_float("color", ui_props.support_beam_color); gpu.state.line_width_set(ui_props.support_beam_width)
        gpu.state.depth_mask_set(True); support_beam_render_batch.draw(beam_render_shader); gpu.state.depth_mask_set(False)

    if hydro_beam_render_batch is not None and ui_props.toggle_hydro_beams_vis:
        beam_render_shader.uniform_float("color", ui_props.hydro_beam_color); gpu.state.line_width_set(ui_props.hydro_beam_width)
        gpu.state.depth_mask_set(True); hydro_beam_render_batch.draw(beam_render_shader); gpu.state.depth_mask_set(False)

    if bounded_beam_render_batch is not None and ui_props.toggle_bounded_beams_vis:
        beam_render_shader.uniform_float("color", ui_props.bounded_beam_color); gpu.state.line_width_set(ui_props.bounded_beam_width)
        gpu.state.depth_mask_set(True); bounded_beam_render_batch.draw(beam_render_shader); gpu.state.depth_mask_set(False)

    if lbeam_render_batch is not None and ui_props.toggle_lbeam_beams_vis:
        beam_render_shader.uniform_float("color", ui_props.lbeam_beam_color); gpu.state.line_width_set(ui_props.lbeam_beam_width)
        gpu.state.depth_mask_set(True); lbeam_render_batch.draw(beam_render_shader); gpu.state.depth_mask_set(False)

    if pressured_beam_render_batch is not None and ui_props.toggle_pressured_beams_vis:
        beam_render_shader.uniform_float("color", ui_props.pressured_beam_color); gpu.state.line_width_set(ui_props.pressured_beam_width)
        gpu.state.depth_mask_set(True); pressured_beam_render_batch.draw(beam_render_shader); gpu.state.depth_mask_set(False)

    if torsionbar_render_batch is not None and ui_props.toggle_torsionbars_vis:
        beam_render_shader.uniform_float("color", ui_props.torsionbar_color); gpu.state.line_width_set(ui_props.torsionbar_width)
        gpu.state.depth_mask_set(True); torsionbar_render_batch.draw(beam_render_shader); gpu.state.depth_mask_set(False)

    if torsionbar_red_render_batch is not None and ui_props.toggle_torsionbars_vis:
        beam_render_shader.uniform_float("color", ui_props.torsionbar_mid_color); gpu.state.line_width_set(ui_props.torsionbar_width)
        gpu.state.depth_mask_set(True); torsionbar_red_render_batch.draw(beam_render_shader); gpu.state.depth_mask_set(False)

    # Rail Draw Call
    if rail_render_batch is not None and ui_props.toggle_rails_vis:
        beam_render_shader.uniform_float("color", ui_props.rail_color); gpu.state.line_width_set(ui_props.rail_width)
        gpu.state.depth_mask_set(True); rail_render_batch.draw(beam_render_shader); gpu.state.depth_mask_set(False)

    # Draw cross-part beams only if toggle is ON
    if cross_part_beam_render_batch is not None and ui_props.toggle_cross_part_beams_vis:
        beam_render_shader.uniform_float("color", ui_props.cross_part_beam_color); gpu.state.line_width_set(ui_props.cross_part_beam_width)
        gpu.state.depth_mask_set(True); cross_part_beam_render_batch.draw(beam_render_shader); gpu.state.depth_mask_set(False)

    # Draw highlight batches
    # ... (highlight drawing remains the same) ...
    highlight_width = jb_globals.highlighted_element_original_width * ui_props.highlight_thickness_multiplier
    gpu.state.depth_mask_set(True) # Ensure highlight draws over others if needed

    if highlight_torsionbar_outer_batch is not None:
        beam_render_shader.uniform_float("color", jb_globals.highlighted_element_color) # Use default outer color
        gpu.state.line_width_set(highlight_width)
        highlight_torsionbar_outer_batch.draw(beam_render_shader)

    if highlight_torsionbar_mid_batch is not None:
        beam_render_shader.uniform_float("color", jb_globals.highlighted_element_mid_color) # Use default mid color
        gpu.state.line_width_set(highlight_width)
        highlight_torsionbar_mid_batch.draw(beam_render_shader)

    if highlight_render_batch is not None: # Draw non-torsionbar highlight
        beam_render_shader.uniform_float("color", jb_globals.highlighted_element_color)
        gpu.state.line_width_set(highlight_width)
        highlight_render_batch.draw(beam_render_shader)

    gpu.state.depth_mask_set(False) # Reset depth mask
    gpu.state.line_width_set(1.0)
